################################################################################
#################              ProDSP Module Tests             #################
################################################################################

import os
import sys
import pandas as pd
import numpy as np
# from aaindex import AAIndex
import pySAR.aaindex as aaindex
import pySAR.proDSP as proDSP_
import pySAR.pySAR as pysar
# from proDSP import ProDSP
# from pySAR import PySAR
import pySAR
import unittest
import requests
import urllib.request

class ProDSPTests(unittest.TestCase):

    def setUp(self):
        """ Import the 4 test datasets used for testing the proDSP methods. """

        try:
            self.test_dataset1 = pd.read_csv(os.path.join('tests','test_data','test_thermostability.txt'),sep=",", header=0)
        except:
            raise IOError('Error reading in test_dataset1.')
        try:
            self.test_dataset2 = pd.read_csv(os.path.join('tests','test_data','test_enantioselectivity.txt'),sep=",", header=0)
        except:
            raise IOError('Error reading in test_dataset2')
        try:
            self.test_dataset3 = pd.read_csv(os.path.join('tests','test_data','test_localization.txt'),sep=",", header=0)
        except:
            raise IOError('Error reading in test_dataset3')
        try:
            self.test_dataset4 = pd.read_csv(os.path.join('tests','test_data','test_absorption.txt'),sep=",", header=0)
        except:
            raise IOError('Error reading in test_dataset4')

        self.aaindex = aaindex.AAIndex()

        self.pySAR = pysar.PySAR(algorithm="PLSReg",dataset=os.path.join('tests','test_data','test_enantioselectivity.txt'), activity="e-value")

    def test_proDSP(self):
        #test general stuff like the input parameters etc

        test_aaindices1 = "BURA740101"
        test_aaindices2 = "COHE430101"
        test_aaindices3 = ["FAUJ880108","NAKH900102","YUTK870103"]
#1.)
        encoded_seq1 = self.pySAR.get_aai_enoding(test_aaindices1)
        proDsp = proDSP_.ProDSP(encoded_seq1)
        self.assertEqual(proDsp.spectrum, "power")
        self.assertEqual(proDsp.window_type, "hamming")
        self.assertIsInstance(proDsp.window, np.ndarray)
        self.assertIsNone(proDsp.filter_)
        self.assertEqual(proDsp.encoded_sequences.shape, (self.pySAR.num_seqs,self.pySAR.seq_len))
        self.assertEqual(proDsp.num_seqs, self.pySAR.num_seqs)
        self.assertEqual(proDsp.signal_len, self.pySAR.seq_len)
        self.assertEqual(proDsp.fft_power.shape, encoded_seq1.shape)
        self.assertEqual(proDsp.fft_real.shape, encoded_seq1.shape)
        self.assertEqual(proDsp.fft_abs.shape, encoded_seq1.shape)
        self.assertEqual(proDsp.fft_imag.shape, encoded_seq1.shape)
        # self.assertTrue(proDsp.spectrum_encoding == proDSP.fft_power)
        self.assertTrue(proDsp.spectrum_encoding.any() == proDsp.fft_power.any())
        self.assertEqual(proDsp.fft_freqs.shape,encoded_seq1.shape)
#2.)
        proDsp = proDSP_.ProDSP(encoded_seq1, window="notawindow")
        self.assertEqual(proDsp.window, 1)
#3.)
        with(self.assertRaises(ValueError)):
            proDsp = proDSP_.ProDSP(encoded_seq1, spectrum="blahblahblah")

        with(self.assertRaises(ValueError)):
            proDsp = proDSP_.ProDSP(encoded_seq1, spectrum=None)
#4.)
        proDsp = proDSP_.ProDSP(encoded_seq1, filter_="blahblahblah")
        self.assertEqual(proDsp.filter_, "")
#5.)
        #test window closeness function
        proDsp = proDSP_.ProDSP(encoded_seq1, window = "hamm")
        self.assertEqual(proDsp.window_type, "hamming")
        proDsp = proDSP_.ProDSP(encoded_seq1, window = "bart")
        self.assertEqual(proDsp.window_type, "bartlett")
        proDsp = proDSP_.ProDSP(encoded_seq1, window = "gausi")
        self.assertEqual(proDsp.window_type, "gaussian")
#6.)
        proDSP = proDSP_.ProDSP(encoded_seq1, spectrum = "absolute")
        self.assertEqual(proDSP.spectrum, "absolute")
        proDSP = proDSP_.ProDSP(encoded_seq1, spectrum = "imaginary")
        self.assertEqual(proDSP.spectrum, "imaginary")

        self.assertEqual(proDsp.window_type, "gaussian")
#7.)
        proDSP = proDSP_.ProDSP(encoded_seq1, filter_ = "savgol")
        self.assertEqual(proDSP.filter_, "savgol")

    def test_preprocessing(self):
        """ Testing preprocessing functionality of proDSP class. """

        test_aaindices1 = "COHE430101"
#1.)
        encoded_seq1 = self.pySAR.get_aai_enoding(test_aaindices1)
        proDsp = proDSP_.ProDSP(encoded_seq1)
        proDsp.pre_processing()
        self.assertTrue(np.all((proDsp.fft_power==0)))
        self.assertTrue(np.all((proDsp.fft_real==0)))
        self.assertTrue(np.all((proDsp.fft_imag==0)))
        self.assertTrue(np.all((proDsp.fft_abs==0)))
        self.assertFalse(np.isnan(proDsp.encoded_sequences).any(),
            'Sequences contain null values.')
#2.)

    def test_protein_spectra(self):
        """ Testing getting protein spectra from encoded protein sequences. """

        test_aaindices1 = "COHE430101"
#1.)
        encoded_seq1 = self.pySAR.get_aai_enoding(test_aaindices1)
        proDsp = proDSP_.ProDSP(encoded_seq1)
        self.assertTrue(proDsp.fft_power.dtype, "complex128")
        self.assertTrue(proDsp.fft_real.dtype, "complex128")
        self.assertTrue(proDsp.fft_imag.dtype, "complex128")
        self.assertTrue(proDsp.fft_abs.dtype, "complex128")

    def test_max_freq(self):
        """ Testing max frequency functionality. """

#1.)
        test_aaindices1 = "COHE430101"
        encoded_seq1 = self.pySAR.get_aai_enoding(test_aaindices1)
        proDsp = proDSP_.ProDSP(encoded_seq1)

# proDSP.fft.dtype == dtype('complex128')
# proDSP.rfft.dtype == dtype('complex128')



################################################################################
#################            Descriptors Module Tests          #################
################################################################################

import pandas as pd
import numpy as np
import os
import unittest
unittest.TestLoader.sortTestMethodsUsing = None
import random

from pySAR.PyBioMed.PyBioMed.PyProtein import AAComposition, Autocorrelation, CTD, ConjointTriad, QuasiSequenceOrder, PseudoAAC
import pySAR.descriptors as descr
import pySAR.utils as utils

class DescriptorTests(unittest.TestCase):

    def setUp(self):
        """ Import the 4 test datasets used for testing the descriptor methods. """

        try:
            self.test_dataset1 = pd.read_csv(os.path.join('tests','test_data','test_thermostability.txt'),sep=",", header=0)
        except:
            raise IOError('Error reading in test_dataset1.')
        try:
            self.test_dataset2 = pd.read_csv(os.path.join('tests','test_data','test_enantioselectivity.txt'),sep=",", header=0)
        except:
            raise IOError('Error reading in test_dataset2.')
        try:
            self.test_dataset3 = pd.read_csv(os.path.join('tests','test_data','test_localization.txt'),sep=",", header=0)
        except:
            raise IOError('Error reading in test_dataset3.')
        try:
            self.test_dataset4 = pd.read_csv(os.path.join('tests','test_data','test_absorption.txt'),sep=",", header=0)
        except:
            raise IOError('Error reading in test_dataset4.')

        #append all datasets to a list
        self.all_test_datasets = [self.test_dataset1, self.test_dataset2, self.test_dataset3,
                self.test_dataset4]

    def test_descriptor(self):
        """ Test descriptor initialisation process. Verify the initial input parameters
        and descriptor attributes are correct. """

        #testing on all 4 datasets
        for dataset in range(0,len(self.all_test_datasets)):
            desc = descr.Descriptors(self.all_test_datasets[dataset]['sequence'], desc_dataset="")
        # desc = Descriptors(self.test_dataset1['sequence'], desc_dataset="")
#1.)
            #verify num_seqs descriptors attribute is correct
            # self.assertEqual(desc.num_seqs, len(self.test_dataset1['sequence']),
            self.assertEqual(desc.num_seqs, len(self.all_test_datasets[dataset]['sequence']),
                'num_seqs attribute not equal to the number of sequences found.')

            #verify all_desc attribute is at its default of False
            self.assertFalse(desc.all_desc, 'all_desc parameter should initially be False.')

            #verify that all input sequences dont have any gaps/missing amino acids
            for seq in desc.protein_seqs:
                self.assertNotIn('-',seq, 'There should be no gaps (-) in the sequences.')
#2.)
            #verify all descriptor attributes are initialised to empty dataframes
            self.assertTrue(desc.aa_composition.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.dipeptide_composition.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.tripeptide_composition.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.normalized_moreaubroto_autocorrelation.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.moran_autocorrelation.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.geary_autocorrelation.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.ctd.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.composition.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.transition.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.distribution.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.conjoint_triad.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.seq_order_coupling_number.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.quasi_seq_order.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.pseudo_aa_composition.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.amp_pseudo_aa_composition.empty, 'Attribute should be initialised to an empty dataframe')
            self.assertTrue(desc.all_descriptors.empty, 'Attribute should be initialised to an empty dataframe')
#3.)
            invalid_testdata1_copy = self.test_dataset1

            #descriptor module should raise ValueError when the full dataset DF is
            #   passed in instead of just the sequences.
            with self.assertRaises(ValueError):
                fail_desc = descr.Descriptors(invalid_testdata1_copy)

            #J,O,U and Z are all invalid amino acid values
            invalid_testdata1_copy['sequence'][0] = 'JOUZ'

            #descriptor module should raise ValueError if invalid sequence found in dataset
            with self.assertRaises(ValueError):
                fail_desc = descr.Descriptors(invalid_testdata1_copy['sequence'])

    def test_descriptor_groups(self):
        """ Testing the descriptor groups dictionary which stores the specific group
        that a descriptor attribute is a member of. """

        #testing on all 4 datasets
        for dataset in range(0,len(self.all_test_datasets)):
            desc = descr.Descriptors(self.all_test_datasets[dataset]['sequence'], desc_dataset="")
#1.)
            self.assertEqual(list(desc.descriptor_groups.keys()), desc.all_descriptors_list())
            self.assertEqual(list(desc.descriptor_groups.values()).count("Composition"), 3)
            self.assertEqual(list(desc.descriptor_groups.values()).count("Autocorrelation"), 3)
            self.assertEqual(list(desc.descriptor_groups.values()).count("Conjoint Triad"), 1)
            self.assertEqual(list(desc.descriptor_groups.values()).count("Quasi-Sequence-Order"), 2)
            self.assertEqual(list(desc.descriptor_groups.values()).count("CTD"), 4)
            self.assertEqual(list(desc.descriptor_groups.values()).count("Pseudo Composition"), 2)
            self.assertEqual(len(desc.descriptor_groups.keys()), len(desc.all_descriptors_list()))
#2.)
            self.assertEqual(desc.descriptor_groups['aa_composition'], "Composition")
            self.assertEqual(desc.descriptor_groups['dipeptide_composition'], "Composition")
            self.assertEqual(desc.descriptor_groups['moran_autocorrelation'], "Autocorrelation")
            self.assertEqual(desc.descriptor_groups['geary_autocorrelation'], "Autocorrelation")
            self.assertEqual(desc.descriptor_groups['composition'], "CTD")
            self.assertEqual(desc.descriptor_groups['distribution'], "CTD")
            self.assertEqual(desc.descriptor_groups['transition'], "CTD")
            self.assertEqual(desc.descriptor_groups['pseudo_aa_composition'], "Pseudo Composition")
            self.assertEqual(desc.descriptor_groups['quasi_seq_order'], "Quasi-Sequence-Order")
            self.assertEqual(desc.descriptor_groups['amp_pseudo_aa_composition'], "Pseudo Composition")
#3.)
            self.assertIsInstance(desc.descriptor_groups, dict)

    def test_aa_composition(self):
        """ Testing Amino Acid Composition protein descriptor attributes and methods. """
#1.)
        #run tests on all test datasets
        for data in range(0,len(self.all_test_datasets)):

            #get random sequence in dataset to calculate descriptor values for
            random_seq = random.randint(0,len(self.all_test_datasets[data])-1)
            desc = descr.Descriptors(self.all_test_datasets[data]['sequence'][random_seq], desc_dataset="")
            aa_comp = desc.get_aa_composition()

            self.assertEqual(aa_comp.shape, (1,20), 'Descriptor not of correct shape.')
            self.assertIsInstance(aa_comp, pd.DataFrame,'Descriptor not of type DataFrame.')
            self.assertEqual(AAComposition.AALetter, list(aa_comp.columns), 'Incorrect column values found.')
            self.assertFalse(aa_comp.empty, 'Descriptor dataframe should not be empty')
            self.assertTrue(aa_comp.any().isnull().sum()== 0,'Descriptor contains null values.')

    def test_dipeptide_composition(self):
        """ Testing Dipeptide Composition protein descriptor attributes and methods. """
#1.)
        #run tests on all test datasets
        for data in range(0,len(self.all_test_datasets)):

            #get random sequence in dataset to calculate descriptor values for
            random_seq = random.randint(0,len(self.all_test_datasets[data])-1)
            desc = descr.Descriptors(self.all_test_datasets[data]['sequence'][random_seq], desc_dataset="")
            dipeptide_comp = desc.get_dipeptide_composition()

            self.assertEqual(dipeptide_comp.shape, (1,400),'Descriptor not of correct shape.')
            self.assertIsInstance(dipeptide_comp, pd.DataFrame,'Descriptor not of type DataFrame.')
            self.assertFalse(dipeptide_comp.empty, 'Descriptor dataframe should not be empty')
            self.assertTrue(dipeptide_comp.any().isnull().sum()== 0,'Descriptor contains null values.')

    # @unittest.skip("Descriptor can take quite a bit of time to calculate therefore skipping")
    def test_tripeptide_composition(self):
        """ Testing Tripeptide Composition protein descriptor attributes and methods. """
#1.)
        #run tests on all test datasets
        for data in range(0,len(self.all_test_datasets)):

            #get random sequence in dataset to calculate descriptor values for
            random_seq = random.randint(0,len(self.all_test_datasets[data])-1)
            desc = descr.Descriptors(self.all_test_datasets[data]['sequence'][random_seq], desc_dataset="")
            tripeptide_comp = desc.get_tripeptide_composition()

            self.assertEqual(tripeptide_comp.shape, (1,8000),'Descriptor not of correct shape.')
            self.assertIsInstance(tripeptide_comp, pd.DataFrame,'Descriptor not of type DataFrame.')
            self.assertFalse(tripeptide_comp.empty, 'Descriptor dataframe should not be empty')
            self.assertTrue(tripeptide_comp.any().isnull().sum()== 0,'Descriptor contains null values.')

    def test_normalized_moreaubroto_autocorrelation(self):
        """ Testing normalized moreaubroto autocorrelation descriptor attributes and methods. """
#1.)
        #run tests on all test datasets
        for data in range(0,len(self.all_test_datasets)):

            #get random sequence in dataset to calculate descriptor values for
            random_seq = random.randint(0,len(self.all_test_datasets[data])-1)
            desc = descr.Descriptors(self.all_test_datasets[data]['sequence'][random_seq], desc_dataset="")
            norm_moreaubroto = desc.get_norm_moreaubroto_autocorrelation()

            self.assertEqual(norm_moreaubroto.shape, (1,240))
            self.assertIsInstance(norm_moreaubroto, pd.DataFrame)
            self.assertFalse(norm_moreaubroto.empty, 'Descriptor dataframe should not be empty')
            self.assertTrue(norm_moreaubroto.any().isnull().sum()== 0)

    def test_moran_autocorrelation(self):
        """ Testing Moran autocorrelation descriptor attributes and methods. """
#1.)
        #run tests on all test datasets
        for data in range(0,len(self.all_test_datasets)):

            #get random sequence in dataset to calculate descriptor values for
            random_seq = random.randint(0,len(self.all_test_datasets[data])-1)
            desc = descr.Descriptors(self.all_test_datasets[data]['sequence'][random_seq], desc_dataset="")
            moran = desc.get_moran_autocorrelation()

            self.assertEqual(moran.shape, (1,240))
            self.assertIsInstance(moran, pd.DataFrame)
            self.assertFalse(moran.empty, 'Descriptor dataframe should not be empty')
            self.assertTrue(moran.any().isnull().sum()== 0)

    def test_geary_autocorrelation(self):
        """ Testing Geary autocorrelation descriptor attributes and methods. """
#1.)
        #run tests on all test datasets
        for data in range(0,len(self.all_test_datasets)):

            #get random sequence in dataset to calculate descriptor values for
            random_seq = random.randint(0,len(self.all_test_datasets[data])-1)
            desc = descr.Descriptors(self.all_test_datasets[data]['sequence'][random_seq], desc_dataset="")
            geary = desc.get_geary_autocorrelation()

            self.assertEqual(geary.shape, (1,240))
            self.assertIsInstance(geary, pd.DataFrame)
            self.assertFalse(geary.empty, 'Descriptor dataframe should not be empty')
            self.assertTrue(geary.any().isnull().sum()== 0)

    def test_ctd(self):
        """ Testing CTD descriptor attributes and methods. """
#1.)
        #run tests on all test datasets
        for data in range(0,len(self.all_test_datasets)):

            #get random sequence in dataset to calculate descriptor values for
            random_seq = random.randint(0,len(self.all_test_datasets[data])-1)
            desc = descr.Descriptors(self.all_test_datasets[data]['sequence'][random_seq], desc_dataset="")
            ctd = desc.get_ctd()

            self.assertEqual(ctd.shape, (1,147))
            self.assertIsInstance(ctd, pd.DataFrame)
            self.assertFalse(ctd.empty, 'Descriptor dataframe should not be empty')
            self.assertTrue(ctd.any().isnull().sum()== 0)

    def test_conjoint_triad(self):
        """ Testing Conjoint Triad descriptor attributes and methods. """
#1.)
        #run tests on all test datasets
        for data in range(0,len(self.all_test_datasets)):

            #get random sequence in dataset to calculate descriptor values for
            random_seq = random.randint(0,len(self.all_test_datasets[data])-1)
            desc = descr.Descriptors(self.all_test_datasets[data]['sequence'][random_seq], desc_dataset="")
            conjoint_triad = desc.get_conjoint_triad()

            self.assertEqual(conjoint_triad.shape, (1,343), 'Descriptor not of correct shape (1,343)')
            self.assertIsInstance(conjoint_triad, pd.DataFrame,'Descriptor not of type DataFrame')
            self.assertFalse(conjoint_triad.empty, 'Descriptor dataframe should not be empty')
            self.assertTrue(conjoint_triad.any().isnull().sum()== 0,'Descriptor contains null values')

    def test_seq_order_coupling_number(self):
        """ Testing sequence order coupling number descriptor attributes and methods. """
#1.)
        #run tests on all test datasets
        for data in range(0,len(self.all_test_datasets)):

            #get random sequence in dataset to calculate descriptor values for
            random_seq = random.randint(0,len(self.all_test_datasets[data])-1)
            desc = descr.Descriptors(self.all_test_datasets[data]['sequence'][random_seq], desc_dataset="")
            seq_order_coupling_number = desc.get_seq_order_coupling_number()

            self.assertEqual(seq_order_coupling_number.shape, (1,60), 'Descriptor not of correct shape (1,60)')
            self.assertIsInstance(seq_order_coupling_number, pd.DataFrame,'Descriptor not of type DataFrame')
            self.assertFalse(seq_order_coupling_number.empty, 'Descriptor dataframe should not be empty')
            self.assertTrue(seq_order_coupling_number.any().isnull().sum()== 0,'Descriptor contains null values')

    def test_quasi_seq_order(self):
        """ Testing Quasi sequence order descriptor attributes and methods. """
#1.)
        #run tests on all test datasets
        for data in range(0,len(self.all_test_datasets)):

            #get random sequence in dataset to calculate descriptor values for
            random_seq = random.randint(0,len(self.all_test_datasets[data])-1)
            desc = descr.Descriptors(self.all_test_datasets[data]['sequence'][random_seq], desc_dataset="")
            quasi_seq_order = desc.get_quasi_seq_order()

            self.assertEqual(quasi_seq_order.shape, (1,100), 'Descriptor not of correct shape (1,100)')
            self.assertIsInstance(quasi_seq_order, pd.DataFrame,'Descriptor not of type DataFrame')
            self.assertFalse(quasi_seq_order.empty, 'Descriptor dataframe should not be empty')
            self.assertTrue(quasi_seq_order.any().isnull().sum()== 0,'Descriptor contains null values')

    @unittest.skip("Descriptor can take quite a bit of time to calculate therefore skipping")
    def test_pseudo_AAC(self):
        """ Testing Pseudo Amino Acid Composition descriptor attributes and methods. """
#1.)
        #run tests on all test datasets
        for data in range(0,len(self.all_test_datasets)):

            #get random sequence in dataset to calculate descriptor values for
            random_seq = random.randint(0,len(self.all_test_datasets[data])-1)
            desc = descr.Descriptors(self.all_test_datasets[data]['sequence'][random_seq], desc_dataset="")
            pseudo_AAC = desc.get_pseudo_aa_composition()

            self.assertEqual(pseudo_AAC.shape, (1,50), 'Descriptor not of correct shape (1,50)')
            self.assertIsInstance(pseudo_AAC, pd.DataFrame, 'Descriptor not of type DataFrame')
            self.assertFalse(pseudo_AAC.empty, 'Descriptor dataframe should not be empty')
            self.assertTrue(pseudo_AAC.any().isnull().sum()== 0, 'Descriptor contains null values')

    @unittest.skip("Descriptor can take quite a bit of time to calculate therefore skipping")
    def test_amp_pseudo_AAC(self):
        """ Testing Amphiphilic Pseudo Amino Acid Composition descriptor attributes and methods. """
#1.)
        #run tests on all test datasets
        for data in range(0,len(self.all_test_datasets)):

            #get random sequence in dataset to calculate descriptor values for
            random_seq = random.randint(0,len(self.all_test_datasets[data])-1)
            desc = descr.Descriptors(self.all_test_datasets[data]['sequence'][random_seq], desc_dataset="")
            amp_pseudo_AAC = desc.get_amp_pseudo_aa_composition()

            self.assertEqual(amp_pseudo_AAC.shape, (1,80), 'Descriptor not of correct shape (1,80)')
            self.assertIsInstance(amp_pseudo_AAC, pd.DataFrame)
            self.assertFalse(amp_pseudo_AAC.empty)
            self.assertTrue(amp_pseudo_AAC.any().isnull().sum()== 0,'Descriptor contains null values')

    def test_get_descriptor_encoding(self):
        """ Testing get_descriptor_encoding function.  """

        valid_desc = [

        'aa_composition', 'dipeptide_composition', 'tripeptide_composition', \
        'normalized_moreaubroto_autocorrelation','moran_autocorrelation','geary_autocorrelation', \
        'ctd', 'composition', 'transition', 'distribution', 'conjoint_triad', \
        'seq_order_coupling_number','quasi_seq_order'

        ]
#1.)
        #run tests on all test datasets
        for data in range(0,len(self.all_test_datasets)):

            #get random sequence in dataset to calculate descriptor values for
            random_seq = random.randint(0,len(self.all_test_datasets[data])-1)
            desc = descr.Descriptors(self.all_test_datasets[data]['sequence'][random_seq], desc_dataset="")

            #test each descriptor is calculated when string of its name passed to func
            for val_desc in range(0,len(valid_desc)):

                encoded_desc = desc.get_descriptor_encoding(valid_desc[val_desc])
                self.assertIsNotNone(encoded_desc)
                self.assertFalse(encoded_desc.empty)
                self.assertTrue(encoded_desc.shape[0],1)

    def tearDown(self):
        """ Cleanup tests and delete datasets. """
        del self.test_dataset1
        del self.test_dataset2
        del self.test_dataset3
        del self.test_dataset4



# -*- coding: utf-8 -*-
#  Copyright (c) 2016-2017, Zhijiang Yao, Jie Dong and Dongsheng Cao
#  All rights reserved.
#  This file is part of the PyBioMed.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the PyBioMed source tree.
"""
#####################################################################################

The prediction of functional sites (e.g.,methylation) of proteins usually needs to

split the total protein into a set of segments around specific amino acid. Given a

specific window size p, we can obtain all segments of length equal to (2*p+1) very

easily. Note that the output of the method is a list form. You can freely use and

distribute it. If you have any problem, you could contact with us timely.

Authors: Zhijiang Yao and Dongsheng Cao.

Date: 2016.06.04

Email: gadsby@163.com

#####################################################################################

"""

# Core Library modules
import re
import string

AALetter = [
    "A",
    "R",
    "N",
    "D",
    "C",
    "E",
    "Q",
    "G",
    "H",
    "I",
    "L",
    "K",
    "M",
    "F",
    "P",
    "S",
    "T",
    "W",
    "Y",
    "V",
]


#############################################################################################


def GetSubSequence(ProteinSequence, ToAA="S", window=3):
    """
    #######################################################################
    Get all 2*window+1 sub-sequences whose cener is ToAA in a protein.

    Usage:

    result=GetSubSequence(protein,ToAA,window)

    Input:protein is a pure problem sequence.

    ToAA is the central (query point) amino acid in the sub-sequence.

    window is the span.

    result is a list form containing all satisfied sub-sequences.
    #######################################################################
    """

    if ToAA not in AALetter:
        ToAA = ProteinSequence[1]

    Num = len(ProteinSequence)
    seqiter = re.finditer(ToAA, ProteinSequence)
    AAindex = []
    for i in seqiter:
        AAindex.append(i.end())

    result = []
    for i in AAindex:
        if i - window > 0 and Num - i + 1 - window > 0:
            temp = ProteinSequence[i - window - 1 : i + window]
            result.append(temp)

    return result


#############################################################################################
if __name__ == "__main__":
    protein = "ADGCGVGEGTGQGPMCNCMCMKWVYADEDAADLESDSFADEDASLESDSFPWSNQRVFCSFADEDAS"
    subseq = GetSubSequence(protein, ToAA="D", window=10)
    print(subseq)
    print(len(subseq))



################################################################################
#################             AAIndex Module Tests             #################
################################################################################

import os
import sys
import numpy as np
# from aaindex import AAIndex
import pySAR.aaindex as aaindex
import pySAR.globals_ as _globals

import unittest
import requests
import urllib.request

#test whitespace functionality
class AAIndexTests(unittest.TestCase):

    def setUp(self):
        """ Inititalise AAIndex object and test directory variable. """

        self.aaindex = aaindex.AAIndex()
        self.TEST_DIR = "tests/test_data"

    @unittest.skip("Don't want to overload the FTP server each time tests are run.")
    def test_download(self):
        """ Test Case to check that the download functionality works for the required AAI1 file.
        The file will firstly be removed from the test directory and then redownloaded, and
        its presence in the directory will pass the test. """
#1.)
        #if AAI1 present in test dir then remove
        if (os.path.isfile(os.path.join(self.TEST_DIR, 'test_aaindex1'))):
            os.remove(os.path.join(self.TEST_DIR, 'test_aaindex1'))

        #assert that OSError exception raised when erroneous directory input to download func
        with self.assertRaises(OSError):
            self.aaindex.download_aaindex(save_dir="Non-existent-directory")
#2.)
        #download AAI1 to local test directory
        self.aaindex.download_aaindex(save_dir=self.TEST_DIR)

        #verify download functionality has worked as desired
        self.assertTrue(os.path.isfile(os.path.join(self.TEST_DIR, 'aaindex1')),
            'AAI did not download correctly to the {} directory.'.format(self.TEST_DIR))

    @unittest.skip("Don't want to overload the FTP server each time tests are run.")
    def test_url(self):
        """ Test Case to check that the URL endpoints used for downloading the AAI databases
        return a 200 status code. """

        AA_INDEX1_URL_FTP = "https://www.genome.jp/ftp/db/community/aaindex/aaindex1"
        AA_INDEX1_URL_HTTPS = "https://www.genome.jp/ftp/db/community/aaindex/aaindex1"
        AA_INDEX2_URL = "https://www.genome.jp/ftp/db/community/aaindex/aaindex2"
        AA_INDEX3_URL = "https://www.genome.jp/ftp/db/community/aaindex/aaindex3"
        wrong_AA_INDEX_URL = "https://www.genome.jp/ftp/BLAH/BLAH/BLAH/BLAH"
#1.)
        #test URL endpoints for AAINDEX are active and give a 200 status code
        r = requests.get(AA_INDEX1_URL_FTP, allow_redirects = True)
        self.assertEqual(r.status_code, 200, 'URL not returning Status Code 200.')

        r = requests.get(AA_INDEX1_URL_HTTPS, allow_redirects = True)
        self.assertEqual(r.status_code, 200, 'URL not returning Status Code 200.')

        r = requests.get(AA_INDEX2_URL, allow_redirects = True)
        self.assertEqual(r.status_code, 200, 'URL not returning Status Code 200.')

        r = requests.get(AA_INDEX3_URL, allow_redirects = True)
        self.assertEqual(r.status_code, 200, 'URL not returning Status Code 200.')
#2.)
        r = requests.get(wrong_AA_INDEX_URL, allow_redirects = True)
        self.assertEqual(r.status_code, 404, 'Errenous URL, got 404 page not found error.')

        #maybe try requests.mock

    def test_get_amino_acids(self):
        """ Test Case to check that only valid Amino Acids are used within the AAIndex class. """

        valid_amino_acids = ['-', 'A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K',
            'L', 'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y']
        amino_acids = self.aaindex.get_amino_acids()
#1.)
        for aa in amino_acids:
            self.assertIn(aa, valid_amino_acids, 'Amino Acid {} not in valid \
                amino acids: {}.'.format(aa, valid_amino_acids))

    def test_num_features(self):
        """ Test Case to check that the correct number of indices/records are present
        in the AAIndex object. To date, 566 indices are present in the database,
        features may be added to the AAI in time so the test takes this into account. """
#1.)
        self.assertTrue(self.aaindex.get_num_records() >= 566, 'Incorrect number \
            of features found in AAI, wanted 566, got {}.'.format(self.aaindex.get_num_records()))

    def test_feature_size(self):
        """ Test Case to check that the AAI has the correct dimensionality. Check that
        the number of keys in parsed JSON from the AAI is correct. The test takes
        into account if more features are added to the database in the future. """
#1.)
        self.assertTrue(len(self.aaindex.get_record_codes())>=566, 'Incorrect number \
            of features found in AAI, wanted 566, got {}.'.format(len(self.aaindex.get_record_codes())))

    def test_get_record_from_code(self):
        """ Test Case to check that the correct amino acid values are returned for
        the feature/index codes using the get_record_from_code function. """
#1.)
        #initialise test feature codes and their correct amino acid values
        feature1 = 'AURR980103'
        feature1_vals = {'A': 1.05, 'L': 0.96, 'R': 0.81, 'K': 0.97, 'N': 0.91,
            'M': 0.99, 'D': 1.39, 'F': 0.95, 'C': 0.6, 'P': 1.05, 'Q': 0.87,
            'S': 0.96, 'E': 1.11, 'T': 1.03, 'G': 1.26, 'W': 1.06, 'H': 1.43,
            'Y': 0.94, 'I': 0.95, 'V': 0.62, '-': 0}
        feature2 = 'FAUJ880113'
        feature2_vals = {'A': 4.76, 'L': 4.79, 'R': 4.3, 'K': 4.27, 'N': 3.64,
            'M': 4.25, 'D': 5.69, 'F': 4.31, 'C': 3.67, 'P': 0.0, 'Q': 4.54,
            'S': 3.83, 'E': 5.48, 'T': 3.87, 'G': 3.77, 'W': 4.75, 'H': 2.84,
            'Y': 4.3, 'I': 4.81, 'V': 4.86, '-': 0}
        feature3 = 'NAGK730101'
        feature3_vals = {'A': 1.29, 'L': 1.23, 'R': 0.83, 'K': 1.23, 'N': 0.77,
            'M': 1.23, 'D': 1.0, 'F': 1.23, 'C': 0.94, 'P': 0.7, 'Q': 1.1,
            'S': 0.78, 'E': 1.54, 'T': 0.87, 'G': 0.72, 'W': 1.06, 'H': 1.29,
            'Y': 0.63, 'I': 0.94, 'V': 0.97, '-': 0}
        feature4 = 'ROBB760105'
        feature4_vals = {'A': -2.3, 'L': 2.3, 'R': 0.4, 'K': -3.3, 'N': -4.1,
            'M': 2.3, 'D': -4.4, 'F': 2.6, 'C': 4.4, 'P': -1.8, 'Q': 1.2,
            'S': -1.7, 'E': -5.0, 'T': 1.3, 'G': -4.2, 'W': -1.0, 'H': -2.5,
            'Y': 4.0, 'I': 6.7, 'V': 6.8, '-': 0}
#1.)
        #get amino acid values for inputted feature/index codes
        record = self.aaindex.get_record_from_code(feature1)

        self.assertIn('description', list(record.keys()))
        self.assertIn('refs', list(record.keys()))
        self.assertIn('notes', list(record.keys()))
        self.assertIn('values', list(record.keys()))

        self.assertEqual(feature1_vals, record['values'], 'Amino acid values gotten \
            from AAI do not match the desired values.')
        self.assertEqual(record['description'], "Normalized positional residue frequency at helix termini N' (Aurora-Rose, 1998)")
        self.assertEqual(record['notes'],'')
        self.assertTrue(record['refs'].startswith("Aurora, R. and Rose, G."))
#2.)
        record = self.aaindex.get_record_from_code(feature2)

        self.assertIn('description', list(record.keys()))
        self.assertIn('refs', list(record.keys()))
        self.assertIn('notes', list(record.keys()))
        self.assertIn('values', list(record.keys()))

        self.assertEqual(feature2_vals, record['values'], 'Amino acid values gotten \
            from AAI do not match the desired values.')
        self.assertEqual(record['description'], 'pK-a(RCOOH) (Fauchere et al., 1988)')
        self.assertEqual(record['notes'],'')
        self.assertTrue(record['refs'].startswith("Fauchere, J.L., Charton, M"))
#3.)
        record = self.aaindex.get_record_from_code(feature3)

        self.assertIn('description', list(record.keys()))
        self.assertIn('refs', list(record.keys()))
        self.assertIn('notes', list(record.keys()))
        self.assertIn('values', list(record.keys()))

        self.assertEqual(feature3_vals, record['values'], 'Amino acid values gotten \
            from AAI do not match the desired values.')
        self.assertEqual(record['description'], 'Normalized frequency of alpha-helix (Nagano, 1973)')
        self.assertEqual(record['notes'],'')
        self.assertTrue(record['refs'].startswith("Nagano, K."))
#4.)
        record = self.aaindex.get_record_from_code(feature4)
        self.assertIn('description', list(record.keys()))
        self.assertIn('refs', list(record.keys()))
        self.assertIn('notes', list(record.keys()))
        self.assertIn('values', list(record.keys()))

        self.assertEqual(feature4_vals, record['values'], 'Amino acid values gotten \
            from AAI do not match the desired values.')
        self.assertEqual(record['description'], 'Information measure for extended (Robson-Suzuki, 1976)')
        self.assertEqual(record['notes'],'')
        self.assertTrue(record['refs'].startswith("Robson, B. and Suzuki, E. 'Conformational properties"))
#5.)
        #testing value error raised when errenous indices put into function
        feature5 = "ABCDEFGH"
        feature6 = "123456"
        feature7 = "blahblahblah"

        with self.assertRaises(ValueError):
            feature_vals = self.aaindex.get_record_from_code(feature5)

        with self.assertRaises(ValueError):
            feature_vals = self.aaindex.get_record_from_code(feature6)

        with self.assertRaises(ValueError):
            feature_vals = self.aaindex.get_record_from_code(feature7)

    def test_get_record_from_name(self):

        feature1 = 'YUTK870101'
        # Unfolding Gibbs energy in water, pH7.0 (Yutani et al., 1987)
        feature2 = 'ZIMJ680103'
# Polarity (Zimmerman et al., 1968)
        feature3 = 'ROSM880103'
# Loss of Side chain hydropathy by helix formation (Roseman, 1988)

        error_feature1 = 'blahblahblah'
        error_feature2 = 'not a record name'
        error_feature3 = 'also not a record name'


    def test_record_codes(self):
        """ Test Case to check that correct feature/record codes are in the parsed JSON
        of all AAI records. Also testing that random erroneous feature codes are
        not present. """

        #testing actual index codes
        feature1 = 'VHEG790101'
        feature2 = 'PONP800107'
        feature3 = 'OOBM770102'
        feature4 = 'NADH010101'
#1.)
        #testing index codes are in the AAI1
        self.assertIn(feature1, self.aaindex.get_record_codes(), 'Index {} \
            not found in list of available indices.'.format(feature1))
        self.assertIn(feature2, self.aaindex.get_record_codes(), 'Index {} \
            not found in list of available indices'.format(feature2))
        self.assertIn(feature3, self.aaindex.get_record_codes(), 'Index {} \
            not found in list of available indices'.format(feature3))
        self.assertIn(feature4, self.aaindex.get_record_codes(), 'Index {} \
            not found in list of available indices'.format(feature4))
#2.)
        #testing bogus index codes
        feature5 = 'ABC1234'
        feature6 = 'ABC12345'
        feature7 = 'ABC123456'
        feature8 = 'ABC1234567'

        #testing errenous index codes are not in the AAI1
        self.assertNotIn(feature5, self.aaindex.get_record_codes(), 'Index {} \
            erroneously found in list of available indices.'.format(feature5))
        self.assertNotIn(feature6, self.aaindex.get_record_codes(), 'Index {} \
            erroneously found in list of available indices.'.format(feature6))
        self.assertNotIn(feature7, self.aaindex.get_record_codes(), 'Index {} \
            erroneously found in list of available indices.'.format(feature7))
        self.assertNotIn(feature8, self.aaindex.get_record_codes(), 'Index {} \
            erroneously found in list of available indices.'.format(feature8))

    def test_category(self):
        """ Test Case to check that the correct category value is returned for each index. """

        feature1 = 'TANS770109'
        feature2 = 'RADA880107'
        feature3 = 'ZIMJ680103'
        feature4 = 'WOLS870102'
        err_feature1 = "blahblahblah"
        err_feature2 = "12345"
        err_feature3 = 100.902
#1.)
        cat1 = self.aaindex.get_category_from_record(feature1)
        cat2 = self.aaindex.get_category_from_record(feature2)
        cat3 = self.aaindex.get_category_from_record(feature3)
        cat4 = self.aaindex.get_category_from_record(feature4)

        self.assertEqual(cat1, 'sec_struct', 'Incorrect category returned, got {} \
            instead of {} '.format(cat1, 'sec_struct'))
        self.assertEqual(cat2, 'hydrophobic','Incorrect category returned, got {} \
            instead of {} '.format(cat2, 'hydrophobic'))
        self.assertEqual(cat3, 'polar','Incorrect category returned, got {} \
            instead of {} '.format(cat3, 'polar'))
        self.assertEqual(cat4, 'meta','Incorrect category returned, got {} \
            instead of {} '.format(cat4, 'meta'))
#2.)
        with (self.assertRaises(ValueError)):
            cat5 = self.aaindex.get_category_from_record(err_feature1)

        with (self.assertRaises(ValueError)):
            cat6 = self.aaindex.get_category_from_record(err_feature2)

        with (self.assertRaises(TypeError)):
            cat7 = self.aaindex.get_category_from_record(err_feature3)

    def test_aaindex_names(self):
        """ Test Case to check that the correct names/description is returned for each index. """

        names = self.aaindex.get_record_names()

        name1 = names[50]
        name2 = names[140]
        name3 = names[368]
        name4 = names[560]

        self.assertEqual(len(names), self.aaindex.get_num_records(), 'Total number \
            of names in the AAI database should equal the number of records')
        self.assertEqual(name1, 'Frequency of the 3rd residue in turn (Chou-Fasman, 1978b)',
            'Incorrect index name returned, got {} instead of {}'.format(name1,
                'Frequency of the 3rd residue in turn (Chou-Fasman, 1978b)'))
        self.assertTrue(name2.startswith('Average relative probability'),
            'Incorrect index name returned, got {} instead of {}'.format(name2,
                'Average relative probability'))
        self.assertIn('chain reversal', name3, 'String {} not found in index name: \
            {}'.format('chain reversal',name3))
        self.assertTrue(name4.endswith('(Karkbara-Knisley, 2016)'), 'Index does \
            not end with {}, got {}'.format('(Karkbara-Knisley, 2016)', name4))

    def test_aaindex_refs(self):
        """ Test Case to check that the correct references are returned for each index. """

        feature1 = 'VELV850101'
        feature2 = 'QIAN880139'
        feature3 = 'NAKH900112'
        feature4 = 'LEVM760103'
        feature5 = "blah_blah"
        feature6 = 123
#1.)
        ref1 = self.aaindex.get_ref_from_record(feature1)
        ref2 = self.aaindex.get_ref_from_record(feature2)
        ref3 = self.aaindex.get_ref_from_record(feature3)
        ref4 = self.aaindex.get_ref_from_record(feature4)

        self.assertTrue(ref1.startswith('Veljkovic'), 'Incorrect index reference, \
            got {} instead of {} '.format(ref1,'' ))
        self.assertIn("globular proteins using neural network models", ref2)
        self.assertIn("hydrophobicity of amino acid composition of mitochondrial proteins", ref3)
        self.assertTrue(ref4.endswith('Nucleic Acids Res. 28, 374 (2000).'))
#2.)
        with (self.assertRaises(ValueError)):
            ref5 = self.aaindex.get_ref_from_record(feature5)

        with (self.assertRaises(TypeError)):
            ref6 = self.aaindex.get_ref_from_record(feature6)

    def test_aaindex_encoding(self):
        """ Testing amino acids one-hot encoding. """
#1.)
        encoding = self.aaindex.get_amino_acids_encoding()
        self.assertEqual(encoding.shape, (20,20))
        self.assertIsInstance(encoding, np.ndarray)

if __name__ == '__main__':
    #run all unit tests
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-
#  Copyright (c) 2016-2017, Zhijiang Yao, Jie Dong and Dongsheng Cao
#  All rights reserved.
#  This file is part of the PyBioMed.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the PyBioMed source tree.
"""
###############################################################################
This module is used for calculating the conjoint triad features only from the

protein sequence information. You can get 7*7*7=343 features.You can freely

use and distribute it. If you hava any problem, you could contact with us timely!

Reference:

Juwen Shen, Jian Zhang, Xiaomin Luo, Weiliang Zhu, Kunqian Yu, Kaixian Chen,

Yixue Li, Huanliang Jiang. Predicting proten-protein interactions based only

on sequences inforamtion. PNAS. 2007 (104) 4337-4341.

Authors: Zhijiang Yao and Dongsheng Cao.

Date: 2016.06.04

Email: gadsby@163.com

###############################################################################
"""

# Core Library modules
import string

###############################################################################
AALetter = [
    "A",
    "R",
    "N",
    "D",
    "C",
    "E",
    "Q",
    "G",
    "H",
    "I",
    "L",
    "K",
    "M",
    "F",
    "P",
    "S",
    "T",
    "W",
    "Y",
    "V",
]

# a Dipole scale (Debye): -, Dipole<1.0; +, 1.0<Dipole<2.0; ++, 2.0<Dipole<3.0; +++, Dipole>3.0; +'+'+', Dipole>3.0 with opposite orientation.
# b Volume scale (Å3): -, Volume<50; +, Volume> 50.
# c Cys is separated from class 3 because of its ability to form disulfide bonds.

_repmat = {
    1: ["A", "G", "V"],
    2: ["I", "L", "F", "P"],
    3: ["Y", "M", "T", "S"],
    4: ["H", "N", "Q", "W"],
    5: ["R", "K"],
    6: ["D", "E"],
    7: ["C"],
}


###############################################################################


def _Str2Num(proteinsequence):
    """
    translate the amino acid letter into the corresponding class based on the

    given form.

    """
    repmat = {}
    for i in _repmat:
        for j in _repmat[i]:
            repmat[j] = i

    res = proteinsequence
    for i in repmat:
        res = res.replace(i, str(repmat[i]))
    return res


###############################################################################
def CalculateConjointTriad(proteinsequence):
    """
    Calculate the conjoint triad features from protein sequence.

    Useage:

    res = CalculateConjointTriad(protein)

    Input: protein is a pure protein sequence.

    Output is a dict form containing all 343 conjoint triad features.
    """
    res = {}
    proteinnum = _Str2Num(proteinsequence)
    for i in range(8):
        for j in range(8):
            for k in range(8):
                temp = str(i) + str(j) + str(k)
                res[temp] = proteinnum.count(temp)
    return res


###############################################################################

if __name__ == "__main__":
    protein = "ADGCGVGEGTGQGPMCNCMCMKWVYADEDAADLESDSFADEDASLESDSFPWSNQRVFCSFADEDAS"
    print(CalculateConjointTriad(protein))



# -*- coding: utf-8 -*-
#  Copyright (c) 2016-2017, Zhijiang Yao, Jie Dong and Dongsheng Cao
#  All rights reserved.
#  This file is part of the PyBioMed.
#  The contents are covered by the terms of the BSD license
#  which is included in the file license.txt, found at the root
#  of the PyBioMed source tree.
"""
Compute the composition, transition and distribution descriptors based on the
different properties of AADs.

The AADs with the same properties is marked as the same number. You can get 147
descriptors for a given protein sequence.

References
----------
.. [1] Inna Dubchak, Ilya Muchink, Stephen R.Holbrook and Sung-Hou Kim.
       Prediction of protein folding class using global description of amino
       acid sequence. Proc.Natl. Acad.Sci.USA, 1995, 92, 8700-8704.

.. [2] Inna Dubchak, Ilya Muchink, Christopher Mayor, Igor Dralyuk and Sung-Hou
       Kim. Recognition of a Protein Fold in the Context of the SCOP
       classification. Proteins: Structure, Function and
       Genetics, 1999, 35, 401-407.

Authors: Zhijiang Yao and Dongsheng Cao.

Date: 2016.06.04

Email: gadsby@163.com
"""

# Core Library modules
import copy
import math

AALetter = [
    "A",
    "R",
    "N",
    "D",
    "C",
    "E",
    "Q",
    "G",
    "H",
    "I",
    "L",
    "K",
    "M",
    "F",
    "P",
    "S",
    "T",
    "W",
    "Y",
    "V",
]

_Hydrophobicity = {"1": "RKEDQN", "2": "GASTPHY", "3": "CLVIMFW"}
# '1'stand for Polar; '2'stand for Neutral, '3' stand for Hydrophobicity

_NormalizedVDWV = {"1": "GASTPD", "2": "NVEQIL", "3": "MHKFRYW"}
# '1'stand for (0-2.78); '2'stand for (2.95-4.0), '3' stand for (4.03-8.08)

_Polarity = {"1": "LIFWCMVY", "2": "CPNVEQIL", "3": "KMHFRYW"}
# '1'stand for (4.9-6.2); '2'stand for (8.0-9.2), '3' stand for (10.4-13.0)

_Charge = {"1": "KR", "2": "ANCQGHILMFPSTWYV", "3": "DE"}
# '1'stand for Positive; '2'stand for Neutral, '3' stand for Negative

_SecondaryStr = {"1": "EALMQKRH", "2": "VIYCWFT", "3": "GNPSD"}
# '1'stand for Helix; '2'stand for Strand, '3' stand for coil

_SolventAccessibility = {"1": "ALFCGIVW", "2": "RKQEND", "3": "MPSTHY"}
# '1'stand for Buried; '2'stand for Exposed, '3' stand for Intermediate

_Polarizability = {"1": "GASDT", "2": "CPNVEQIL", "3": "KMHFRYW"}
# '1'stand for (0-0.108); '2'stand for (0.128-0.186), '3' stand for (0.219-0.409)


# You can continuely add other properties of AADs to compute descriptors of
# protein sequence.

_AATProperty = (
    _Hydrophobicity,
    _NormalizedVDWV,
    _Polarity,
    _Charge,
    _SecondaryStr,
    _SolventAccessibility,
    _Polarizability,
)

_AATPropertyName = (
    "_Hydrophobicity",
    "_NormalizedVDWV",
    "_Polarity",
    "_Charge",
    "_SecondaryStr",
    "_SolventAccessibility",
    "_Polarizability",
)


def StringtoNum(ProteinSequence, AAProperty):
    """
    Tranform the protein sequence into the string form such as 32123223132121123.

    Usage:

    result=StringtoNum(protein,AAProperty)

    Input: protein is a pure protein sequence.

    AAProperty is a dict form containing classifciation of amino acids such as _Polarizability.

    Output: result is a string such as 123321222132111123222
    """
    hardProteinSequence = copy.deepcopy(ProteinSequence)
    for k, m in AAProperty.items():
        for index in m:
            hardProteinSequence = hardProteinSequence.replace(index, k)
    TProteinSequence = hardProteinSequence

    return TProteinSequence


def CalculateComposition(ProteinSequence, AAProperty, AAPName):
    """
    ###############################################################################################
    A method used for computing composition descriptors.

    Usage:

    result=CalculateComposition(protein,AAProperty,AAPName)

    Input: protein is a pure protein sequence.

    AAProperty is a dict form containing classifciation of amino acids such as _Polarizability.

    AAPName is a string used for indicating a AAP name.

    Output: result is a dict form containing composition descriptors based on the given property.
    ###############################################################################################
    """
    TProteinSequence = StringtoNum(ProteinSequence, AAProperty)
    Result = {}
    Num = len(TProteinSequence)
    Result[AAPName + "C" + "1"] = round(float(TProteinSequence.count("1")) / Num, 3)
    Result[AAPName + "C" + "2"] = round(float(TProteinSequence.count("2")) / Num, 3)
    Result[AAPName + "C" + "3"] = round(float(TProteinSequence.count("3")) / Num, 3)
    return Result


def CalculateTransition(ProteinSequence, AAProperty, AAPName):
    """
    ###############################################################################################
    A method used for computing transition descriptors

    Usage:

    result=CalculateTransition(protein,AAProperty,AAPName)

    Input:protein is a pure protein sequence.

    AAProperty is a dict form containing classifciation of amino acids such as _Polarizability.

    AAPName is a string used for indicating a AAP name.

    Output:result is a dict form containing transition descriptors based on the given property.
    ###############################################################################################
    """

    TProteinSequence = StringtoNum(ProteinSequence, AAProperty)
    Result = {}
    Num = len(TProteinSequence)
    CTD = TProteinSequence
    Result[AAPName + "T" + "12"] = round(
        float(CTD.count("12") + CTD.count("21")) / (Num - 1), 3
    )
    Result[AAPName + "T" + "13"] = round(
        float(CTD.count("13") + CTD.count("31")) / (Num - 1), 3
    )
    Result[AAPName + "T" + "23"] = round(
        float(CTD.count("23") + CTD.count("32")) / (Num - 1), 3
    )
    return Result


def CalculateDistribution(ProteinSequence, AAProperty, AAPName):
    """
    ###############################################################################################
    A method used for computing distribution descriptors.

    Usage:

    result=CalculateDistribution(protein,AAProperty,AAPName)

    Input:protein is a pure protein sequence.

    AAProperty is a dict form containing classifciation of amino acids such as _Polarizability.

    AAPName is a string used for indicating a AAP name.

    Output:result is a dict form containing Distribution descriptors based on the given property.
    ###############################################################################################
    """
    TProteinSequence = StringtoNum(ProteinSequence, AAProperty)
    Result = {}
    Num = len(TProteinSequence)
    for i in ("1", "2", "3"):
        num = TProteinSequence.count(i)
        ink = 1
        indexk = 0
        cds = []
        while ink <= num:
            indexk = TProteinSequence.find(i, indexk) + 1
            cds.append(indexk)
            ink = ink + 1

        if cds == []:
            Result[AAPName + "D" + i + "001"] = 0
            Result[AAPName + "D" + i + "025"] = 0
            Result[AAPName + "D" + i + "050"] = 0
            Result[AAPName + "D" + i + "075"] = 0
            Result[AAPName + "D" + i + "100"] = 0
        else:
            Result[AAPName + "D" + i + "001"] = round(float(cds[0]) / Num * 100, 3)
            Result[AAPName + "D" + i + "025"] = round(
                float(cds[int(math.floor(num * 0.25)) - 1]) / Num * 100, 3
            )
            Result[AAPName + "D" + i + "050"] = round(
                float(cds[int(math.floor(num * 0.5)) - 1]) / Num * 100, 3
            )
            Result[AAPName + "D" + i + "075"] = round(
                float(cds[int(math.floor(num * 0.75)) - 1]) / Num * 100, 3
            )
            Result[AAPName + "D" + i + "100"] = round(float(cds[-1]) / Num * 100, 3)

    return Result


def CalculateCompositionHydrophobicity(ProteinSequence):
    """
    ###############################################################################################
    A method used for calculating composition descriptors based on Hydrophobicity of

    AADs.

    Usage:

    result=CalculateCompositionHydrophobicity(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing Composition descriptors based on Hydrophobicity.
    ###############################################################################################
    """

    result = CalculateComposition(ProteinSequence, _Hydrophobicity, "_Hydrophobicity")
    return result


def CalculateCompositionNormalizedVDWV(ProteinSequence):
    """
    ###############################################################################################
    A method used for calculating composition descriptors based on NormalizedVDWV of

    AADs.

    Usage:

    result=CalculateCompositionNormalizedVDWV(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing Composition descriptors based on NormalizedVDWV.
    ###############################################################################################
    """
    result = CalculateComposition(ProteinSequence, _NormalizedVDWV, "_NormalizedVDWV")
    return result


def CalculateCompositionPolarity(ProteinSequence):
    """
    ###############################################################################################
    A method used for calculating composition descriptors based on Polarity of

    AADs.

    Usage:

    result=CalculateCompositionPolarity(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing Composition descriptors based on Polarity.
    ###############################################################################################
    """

    result = CalculateComposition(ProteinSequence, _Polarity, "_Polarity")
    return result


def CalculateCompositionCharge(ProteinSequence):
    """
    ###############################################################################################
    A method used for calculating composition descriptors based on Charge of

    AADs.

    Usage:

    result=CalculateCompositionCharge(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing Composition descriptors based on Charge.
    ###############################################################################################
    """

    result = CalculateComposition(ProteinSequence, _Charge, "_Charge")
    return result


def CalculateCompositionSecondaryStr(ProteinSequence):
    """
    ###############################################################################################
    A method used for calculating composition descriptors based on SecondaryStr of

    AADs.

    Usage:

    result=CalculateCompositionSecondaryStr(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing Composition descriptors based on SecondaryStr.
    ###############################################################################################
    """

    result = CalculateComposition(ProteinSequence, _SecondaryStr, "_SecondaryStr")
    return result


def CalculateCompositionSolventAccessibility(ProteinSequence):
    """
    ###############################################################################################
    A method used for calculating composition descriptors based on SolventAccessibility

    of  AADs.

    Usage:

    result=CalculateCompositionSolventAccessibility(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing Composition descriptors based on SolventAccessibility.
    ###############################################################################################
    """

    result = CalculateComposition(
        ProteinSequence, _SolventAccessibility, "_SolventAccessibility"
    )
    return result


def CalculateCompositionPolarizability(ProteinSequence):
    """
    ###############################################################################################
    A method used for calculating composition descriptors based on Polarizability of

    AADs.

    Usage:

    result=CalculateCompositionPolarizability(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing Composition descriptors based on Polarizability.
    ###############################################################################################
    """

    result = CalculateComposition(ProteinSequence, _Polarizability, "_Polarizability")
    return result


##################################################################################################


##################################################################################################
def CalculateTransitionHydrophobicity(ProteinSequence):
    """
    ###############################################################################################
    A method used for calculating Transition descriptors based on Hydrophobicity of

    AADs.

    Usage:

    result=CalculateTransitionHydrophobicity(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing Transition descriptors based on Hydrophobicity.
    ###############################################################################################
    """

    result = CalculateTransition(ProteinSequence, _Hydrophobicity, "_Hydrophobicity")
    return result


def CalculateTransitionNormalizedVDWV(ProteinSequence):
    """
    ###############################################################################################
    A method used for calculating Transition descriptors based on NormalizedVDWV of

    AADs.

    Usage:

    result=CalculateTransitionNormalizedVDWV(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing Transition descriptors based on NormalizedVDWV.
    ###############################################################################################
    """

    result = CalculateTransition(ProteinSequence, _NormalizedVDWV, "_NormalizedVDWV")
    return result


def CalculateTransitionPolarity(ProteinSequence):
    """
    ###############################################################################################
    A method used for calculating Transition descriptors based on Polarity of

    AADs.

    Usage:

    result=CalculateTransitionPolarity(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing Transition descriptors based on Polarity.
    ###############################################################################################
    """

    result = CalculateTransition(ProteinSequence, _Polarity, "_Polarity")
    return result


def CalculateTransitionCharge(ProteinSequence):
    """
    ###############################################################################################
    A method used for calculating Transition descriptors based on Charge of

    AADs.

    Usage:

    result=CalculateTransitionCharge(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing Transition descriptors based on Charge.
    ###############################################################################################
    """

    result = CalculateTransition(ProteinSequence, _Charge, "_Charge")
    return result


def CalculateTransitionSecondaryStr(ProteinSequence):
    """
    ###############################################################################################
    A method used for calculating Transition descriptors based on SecondaryStr of

    AADs.

    Usage:

    result=CalculateTransitionSecondaryStr(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing Transition descriptors based on SecondaryStr.
    ###############################################################################################
    """

    result = CalculateTransition(ProteinSequence, _SecondaryStr, "_SecondaryStr")
    return result


def CalculateTransitionSolventAccessibility(ProteinSequence):
    """
    ###############################################################################################
    A method used for calculating Transition descriptors based on SolventAccessibility

    of  AADs.

    Usage:

    result=CalculateTransitionSolventAccessibility(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing Transition descriptors based on SolventAccessibility.
    ###############################################################################################
    """

    result = CalculateTransition(
        ProteinSequence, _SolventAccessibility, "_SolventAccessibility"
    )
    return result


def CalculateTransitionPolarizability(ProteinSequence):
    """
    ###############################################################################################
    A method used for calculating Transition descriptors based on Polarizability of

    AADs.

    Usage:

    result=CalculateTransitionPolarizability(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing Transition descriptors based on Polarizability.
    ###############################################################################################
    """

    result = CalculateTransition(ProteinSequence, _Polarizability, "_Polarizability")
    return result


##################################################################################################
##################################################################################################
def CalculateDistributionHydrophobicity(ProteinSequence):
    """
    ###############################################################################################
    A method used for calculating Distribution descriptors based on Hydrophobicity of

    AADs.

    Usage:

    result=CalculateDistributionHydrophobicity(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing Distribution descriptors based on Hydrophobicity.
    ###############################################################################################
    """

    result = CalculateDistribution(ProteinSequence, _Hydrophobicity, "_Hydrophobicity")
    return result


def CalculateDistributionNormalizedVDWV(ProteinSequence):
    """
    ###############################################################################################
    A method used for calculating Distribution descriptors based on NormalizedVDWV of

    AADs.

    Usage:

    result=CalculateDistributionNormalizedVDWV(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing Distribution descriptors based on NormalizedVDWV.
    ###############################################################################################
    """

    result = CalculateDistribution(ProteinSequence, _NormalizedVDWV, "_NormalizedVDWV")
    return result


def CalculateDistributionPolarity(ProteinSequence):
    """
    ###############################################################################################
    A method used for calculating Distribution descriptors based on Polarity of

    AADs.

    Usage:

    result=CalculateDistributionPolarity(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing Distribution descriptors based on Polarity.
    ###############################################################################################
    """

    result = CalculateDistribution(ProteinSequence, _Polarity, "_Polarity")
    return result


def CalculateDistributionCharge(ProteinSequence):
    """
    ###############################################################################################
    A method used for calculating Distribution descriptors based on Charge of

    AADs.

    Usage:

    result=CalculateDistributionCharge(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing Distribution descriptors based on Charge.
    ###############################################################################################
    """

    result = CalculateDistribution(ProteinSequence, _Charge, "_Charge")
    return result


def CalculateDistributionSecondaryStr(ProteinSequence):
    """
    ###############################################################################################
    A method used for calculating Distribution descriptors based on SecondaryStr of

    AADs.

    Usage:

    result=CalculateDistributionSecondaryStr(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing Distribution descriptors based on SecondaryStr.
    ###############################################################################################
    """

    result = CalculateDistribution(ProteinSequence, _SecondaryStr, "_SecondaryStr")
    return result


def CalculateDistributionSolventAccessibility(ProteinSequence):
    """
    ###############################################################################################
    A method used for calculating Distribution descriptors based on SolventAccessibility

    of  AADs.

    Usage:

    result=CalculateDistributionSolventAccessibility(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing Distribution descriptors based on SolventAccessibility.
    ###############################################################################################
    """

    result = CalculateDistribution(
        ProteinSequence, _SolventAccessibility, "_SolventAccessibility"
    )
    return result


def CalculateDistributionPolarizability(ProteinSequence):
    """
    ###############################################################################################
    A method used for calculating Distribution descriptors based on Polarizability of

    AADs.

    Usage:

    result=CalculateDistributionPolarizability(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing Distribution descriptors based on Polarizability.
    ###############################################################################################
    """

    result = CalculateDistribution(ProteinSequence, _Polarizability, "_Polarizability")
    return result


##################################################################################################


def CalculateC(ProteinSequence):
    """
    ###############################################################################################
    Calculate all composition descriptors based seven different properties of AADs.

    Usage:

    result=CalculateC(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing all composition descriptors.
    ###############################################################################################
    """
    result = {}
    result.update(CalculateCompositionPolarizability(ProteinSequence))
    result.update(CalculateCompositionSolventAccessibility(ProteinSequence))
    result.update(CalculateCompositionSecondaryStr(ProteinSequence))
    result.update(CalculateCompositionCharge(ProteinSequence))
    result.update(CalculateCompositionPolarity(ProteinSequence))
    result.update(CalculateCompositionNormalizedVDWV(ProteinSequence))
    result.update(CalculateCompositionHydrophobicity(ProteinSequence))
    return result


def CalculateT(ProteinSequence):
    """
    ###############################################################################################
    Calculate all transition descriptors based seven different properties of AADs.

    Usage:

    result=CalculateT(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing all transition descriptors.
    ###############################################################################################
    """
    result = {}
    result.update(CalculateTransitionPolarizability(ProteinSequence))
    result.update(CalculateTransitionSolventAccessibility(ProteinSequence))
    result.update(CalculateTransitionSecondaryStr(ProteinSequence))
    result.update(CalculateTransitionCharge(ProteinSequence))
    result.update(CalculateTransitionPolarity(ProteinSequence))
    result.update(CalculateTransitionNormalizedVDWV(ProteinSequence))
    result.update(CalculateTransitionHydrophobicity(ProteinSequence))
    return result


def CalculateD(ProteinSequence):
    """
    ###############################################################################################
    Calculate all distribution descriptors based seven different properties of AADs.

    Usage:

    result=CalculateD(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing all distribution descriptors.
    ###############################################################################################
    """
    result = {}
    result.update(CalculateDistributionPolarizability(ProteinSequence))
    result.update(CalculateDistributionSolventAccessibility(ProteinSequence))
    result.update(CalculateDistributionSecondaryStr(ProteinSequence))
    result.update(CalculateDistributionCharge(ProteinSequence))
    result.update(CalculateDistributionPolarity(ProteinSequence))
    result.update(CalculateDistributionNormalizedVDWV(ProteinSequence))
    result.update(CalculateDistributionHydrophobicity(ProteinSequence))
    return result


def CalculateCTD(ProteinSequence):
    """
    ###############################################################################################
    Calculate all CTD descriptors based seven different properties of AADs.

    Usage:

    result=CalculateCTD(protein)

    Input:protein is a pure protein sequence.

    Output:result is a dict form containing all CTD descriptors.
    ###############################################################################################
    """
    result = {}
    result.update(CalculateCompositionPolarizability(ProteinSequence))
    result.update(CalculateCompositionSolventAccessibility(ProteinSequence))
    result.update(CalculateCompositionSecondaryStr(ProteinSequence))
    result.update(CalculateCompositionCharge(ProteinSequence))
    result.update(CalculateCompositionPolarity(ProteinSequence))
    result.update(CalculateCompositionNormalizedVDWV(ProteinSequence))
    result.update(CalculateCompositionHydrophobicity(ProteinSequence))
    result.update(CalculateTransitionPolarizability(ProteinSequence))
    result.update(CalculateTransitionSolventAccessibility(ProteinSequence))
    result.update(CalculateTransitionSecondaryStr(ProteinSequence))
    result.update(CalculateTransitionCharge(ProteinSequence))
    result.update(CalculateTransitionPolarity(ProteinSequence))
    result.update(CalculateTransitionNormalizedVDWV(ProteinSequence))
    result.update(CalculateTransitionHydrophobicity(ProteinSequence))
    result.update(CalculateDistributionPolarizability(ProteinSequence))
    result.update(CalculateDistributionSolventAccessibility(ProteinSequence))
    result.update(CalculateDistributionSecondaryStr(ProteinSequence))
    result.update(CalculateDistributionCharge(ProteinSequence))
    result.update(CalculateDistributionPolarity(ProteinSequence))
    result.update(CalculateDistributionNormalizedVDWV(ProteinSequence))
    result.update(CalculateDistributionHydrophobicity(ProteinSequence))
    return result


##################################################################################################

if __name__ == "__main__":
    protein = "ADGCGVGEGTGQGPMCNCMCMKWVYADEDAADLESDSFADEDASLESDSFPWSNQRVFCSFADEDAS"
    print(CalculateCTD(protein))




################################################################################
#################                  Descriptors                 #################
################################################################################

import pandas as pd
import numpy as np
import datetime, time
import itertools
import pickle
import os
import time
from difflib import get_close_matches
import json
import sys
from tqdm import tqdm
from sklearn.preprocessing import StandardScaler

from .globals_ import OUTPUT_DIR, OUTPUT_FOLDER, DATA_DIR
from .aaindex import  AAIndex
from .model import Model
from .proDSP import ProDSP
from .evaluate import Evaluate
from .pySAR import *
from .utils import *
from .PyBioMed.PyBioMed.PyProtein import AAComposition, Autocorrelation, CTD, \
    ConjointTriad, QuasiSequenceOrder, PseudoAAC

class Descriptors():
    """
    Class for calculating a wide variety of protein descriptors. These descriptors
    have been used in a wide variety of Bioinformaitcs applications including
    protein strucutral and function class prediction, protein-protein interactions,
    subcellular location, secondary structure prediction etc. They represent the different
    structural, functional & interaction profiles of proteins by exploring the
    features in the groups of composition, correlation, distribution of the constituent
    residues and their biochemical and physiochemical properties.

    This class allows calculation of the following descriptors: amino-acid compostion (AAComp),
    dipeptide composition (DPComp), tripeptide composition (TPComp), Normalized Moreau-Broto
    autocorrelation (NMBAuto), Moran Autocorrelation (MAuto), Geary Autocorrelation (GAuto),
    Composition (C), Transition (T), Distribution (D), CTD, Conjoint Triad (CTriad), sequence
    order coupling number (SOCNum), Quasi-sequence-order (QSOrder), Pseudo amino-acid
    composition-type 1(PAAcomp) and Amphiphilic pseudo amino-acid composition -type 2 (APAAComp).

    Descriptors are calculated using the PyProtein module from the PyBioMed Python
    package:
    https://pybiomed.readthedocs.io/en/latest/reference/PyProtein.html
    "Copyright (c) 2016-2017, Zhijiang Yao, Jie Dong and Dongsheng Cao
    All rights reserved"

    The output of the descriptor functions in PyBioMed are incompatible with some
    modules and functions in pySAR, therefore the output from PyBioMed is
    postprocessed and reformatted into dataframes to make it useable with pySAR.

    Attributes
    ----------
    protein_seqs : np.ndarray
        array of protein sequences that descriptors will be calculated for.
    desc_dataset : str (default = "descriptors.csv")
        csv file storing pre-calculated descriptor values for a particular
        dataset of protein sequences. By default the class will look for a file
        named 'descriptors.csv' in the DATA_DIR & import the pre-calculated descriptor
        values. If file not found then descriptor values will have to be calculated. It is
        reccomended to pre-calculate the descriptor values to save time and resources
        calculating each time.
    all_desc : bool (default = False)
        if set to true, upon instantiation of a Descriptor class object, all
        available descriptors will be calculated and saved into an output csv.
        This should only have to be done once to calculate all descriptors in 1 go.

    Methods
    -------
    import_descriptors(descriptor_file)
    get_aa_composition()
    get_dipeptide_composition()
    get_tripeptide_composition()
    get_norm_moreaubroto_autocorrelation()
    ......

    References
    ----------
    [1]: Dong, J., Yao, ZJ., Zhang, L. et al. PyBioMed: a python library for
         various molecular representations of chemicals, proteins and DNAs and
         their interactions. J Cheminform 10, 16 (2018).
         https://doi.org/10.1186/s13321-018-0270-2
    [2]: Reczko, M. and Bohr, H. (1994) The DEF data base of sequence based protein
         fold class predictions. Nucleic Acids Res, 22, 3616-3619.
    [3]: Hua, S. and Sun, Z. (2001) Support vector machine approach for protein
         subcellular localization prediction. Bioinformatics, 17, 721-728.
    [4]: Broto P, Moreau G, Vandicke C: Molecular structures: perception,
         autocorrelation descriptor and SAR studies. Eur J Med Chem 1984, 19: 71–78.
    [5]: Ong, S.A., Lin, H.H., Chen, Y.Z. et al. Efficacy of different protein
         descriptors in predicting protein functional families. BMC Bioinformatics
         8, 300 (2007). https://doi.org/10.1186/1471-2105-8-300
    [6]: Inna Dubchak, Ilya Muchink, Stephen R.Holbrook and Sung-Hou Kim. Prediction
         of protein folding class using global description of amino acid sequence.
         Proc.Natl. Acad.Sci.USA, 1995, 92, 8700-8704.
    [7]: Juwen Shen, Jian Zhang, Xiaomin Luo, Weiliang Zhu, Kunqian Yu, Kaixian Chen,
         Yixue Li, Huanliang Jiang. Predicting proten-protein interactions based only
         on sequences inforamtion. PNAS. 2007 (104) 4337-4341.
    [8]: Kuo-Chen Chou. Prediction of Protein Subcellar Locations by Incorporating
         Quasi-Sequence-Order Effect. Biochemical and Biophysical Research
         Communications 2000, 278, 477-483.<- quasi-seq-order refernece
    [9]: Kuo-Chen Chou. Prediction of Protein Cellular Attributes Using
         Pseudo-Amino Acid Composition. PROTEINS: Structure, Function, and
         Genetics, 2001, 43: 246-255.
    [10]: Kuo-Chen Chou. Using amphiphilic pseudo amino acid composition to predict enzyme
          subfamily classes. Bioinformatics, 2005,21,10-19.
    """
    def __init__(self, protein_seqs, desc_dataset="descriptors.csv", all_desc=False):

        self.protein_seqs = protein_seqs
        self.desc_dataset = desc_dataset
        self.all_desc = all_desc

        #if 1 protein sequence (1 string) input then convert to pandas Series object
        if isinstance(self.protein_seqs, str):
          self.protein_seqs = pd.Series(self.protein_seqs)

        #only the sequences should be passed in, not all columns in a dataset etc.
        if isinstance(self.protein_seqs, pd.DataFrame) and \
            len(self.protein_seqs.columns) > 1:
            raise ValueError('The full dataset must not be passed in, only the \
                columns containing the protein sequences.')

        #remove any gaps from protein sequences
        self.protein_seqs = remove_gaps(self.protein_seqs)

        #validate that all inputtted protein sequences are valid and only contain
        #   valid amino acids, if not then raise ValueError
        invalid_seqs = valid_sequence(self.protein_seqs)
        if invalid_seqs!=None:
            raise ValueError('Invalid Amino Acids found in protein sequence \
                dataset: {}'.format(invalid_seqs))

        #get the total number of inputted protein sequences
        self.num_seqs = len(self.protein_seqs)

        #initialise all descriptor attributes to empty dataframes
        self.aa_composition = pd.DataFrame()
        self.dipeptide_composition = pd.DataFrame()
        self.tripeptide_composition = pd.DataFrame()
        self.normalized_moreaubroto_autocorrelation = pd.DataFrame()
        self.moran_autocorrelation = pd.DataFrame()
        self.geary_autocorrelation = pd.DataFrame()
        self.ctd = pd.DataFrame()
        self.composition = pd.DataFrame()
        self.transition = pd.DataFrame()
        self.distribution = pd.DataFrame()
        self.conjoint_triad = pd.DataFrame()
        self.seq_order_coupling_number = pd.DataFrame()
        self.quasi_seq_order = pd.DataFrame()
        self.pseudo_aa_composition = pd.DataFrame()
        self.amp_pseudo_aa_composition = pd.DataFrame()
        self.all_descriptors = pd.DataFrame()

        #try importing descriptors csv with pre-calculated descriptor values,
        #   if not found then calculate all descriptors if all_desc is true
        self.desc_dataset = os.path.join(DATA_DIR, desc_dataset)
        if os.path.isfile(self.desc_dataset):
            self.import_descriptors(self.desc_dataset)
        else:
            #if all_desc parameter true then calculate all descriptor value and store
            #   in their respective attributes
            if all_desc:
                self.all_descriptors = self.get_all_descriptors()
                #save all calculated descriptor values for next time
                self.all_descriptors.to_csv(os.path.join(DATA_DIR, self.desc_dataset),index=0)

        #create dictionary of descriptors and their associated groups
        keys = self.all_descriptors_list()
        values = ["Composition"]*3 + ["Autocorrelation"]*3 + ["CTD"]*4 + ["Conjoint Triad"] + \
            ["Quasi-Sequence-Order"]*2 + ["Pseudo Composition"]*2
        self.descriptor_groups = dict(zip(keys,values))

    def import_descriptors(self, descriptor_file):
        """
        By default, the class will search for a file in the DATA_DIR called
        descriptors.csv with ALL of the pre-calculated descriptor values.
        This function parses this file and sets the descriptor instance variables
        to the correct feature values. The descriptors file should be of size N x 9920,
        where N is the number of protein sequences in the dataset and 9920 is the
        total number of features that can be calculated from the 15 descriptors.
        The method does not work if the file is not of this size.
        To create a descriptors file, create an instance of the desc class,
        with the parameter all_desc=True; this will calculate all descriptors
        and store them to the DATA_DIR in a csv.

        Parameters
        ----------
        descriptor_file : str
            file name containing pre-calculated descriptor values. Ideally, the
            file should be stored in the DATA_DIR.
        """
        # print('Importing Descriptors File....')

        try:
            descriptor_df = pd.read_csv(descriptor_file)
        except IOError:
            print('Error reading descriptor file: {}.'.format(descriptor_file))

        if (descriptor_df.shape)!=(self.num_seqs, 9920):
            raise ValueError('Descriptors file must be of the shape ({}x9920) such \
                that all descriptor values can be parsed from the file.'.format(self.num_seqs))

        #replacing any +/- infinity or NAN values with 0
        descriptor_df.replace([np.inf, -np.inf], np.nan)
        descriptor_df = descriptor_df.fillna(0)

        #get the feature columns for all descriptors
        self.aa_composition = descriptor_df.iloc[:,: 20]
        self.dipeptide_composition = descriptor_df.iloc[:,20:420]
        self.tripeptide_composition = descriptor_df.iloc[:,420:8420]
        self.normalized_moreaubroto_autocorrelation = descriptor_df.iloc[:,8420:8660]
        self.moran_autocorrelation = descriptor_df.iloc[:,8660: 8900]
        self.geary_autocorrelation = descriptor_df.iloc[:,8900:9140]
        self.ctd =  descriptor_df.iloc[:,9140:9287]
        self.composition = descriptor_df.iloc[:,9140:9161]
        self.transition = descriptor_df.iloc[:,9161:9182]
        self.distribution = descriptor_df.iloc[:,9182:9287]
        self.conjoint_triad = descriptor_df.iloc[:,9287:9630]
        self.seq_order_coupling_number = descriptor_df.iloc[:,9630:9690]
        self.quasi_seq_order = descriptor_df.iloc[:,9690:9790]
        self.pseudo_aa_composition = descriptor_df.iloc[:,9790:9840]
        self.amp_pseudo_aa_composition = descriptor_df.iloc[:,9840:9920]
        self.all_descriptors = descriptor_df.iloc[:,:]

    def get_aa_composition(self):
        """
        Calculate Amino Acid Composition (AAComp) of protein sequences. AAComp
        describes the fraction of each amino acid type within a protein sequence,
        and is calculated as:

        AA_Comp(s) = AA(t)/N(s)

        where AA_Comp(s) is the AAComp of protein sequence s, AA(t) is the number
        of amino acid types t (where t = 1,2,..,20) and N(s) is the length of the
        sequence s.

        Returns
        -------
        aa_comp_df : pd.DataFrame
            dataframe of AAComp for all protein sequences. DataFrame will
            be of the shape N x 20, where N is the number of protein sequences and
            20 is the number of features calculated from the descriptor (for the
            20 amino acids).
        """
        print('\nGetting AA Composition Descriptors...')
        print('#####################################\n')

        #if attribute already calculated & not empty then return it
        if not self.aa_composition.empty:
            return self.aa_composition

        aa_comp = []

        #get feature names for AAComp descriptor
        aa = list((AAComposition.CalculateAAComposition(self.protein_seqs[0])).keys())

        #iterate through all sequences, calculate descriptor values and append to aa_comp list
        for seq in tqdm(self.protein_seqs,unit=" sequences",position=0,desc="AA Composition",file=sys.stdout):
            AAComp = AAComposition.CalculateAAComposition(seq)
            aa_comp.append(list(AAComp.values()))

        #convert calculated AAComp values into dataframe
        aa_comp_df = pd.DataFrame(data=aa_comp, columns=aa)

        self.aa_composition = aa_comp_df     #set descriptor attribute

        return aa_comp_df

    def get_dipeptide_composition(self):
        """
        Calculate Dipeptide Composition (DPComp) of protein sequences.
        Dipeptide composition is the fraction of each dipeptide type within a
        protein sequence. With dipeptides being of length 2 and there being 20
        canonical amino acids this creates 20^2 different combinations, thus a
        400-Dimensional vector will be produced such that:

        DPComp(s,t) = AA(s,t) / N -1

        where DPComp(s,t) is the dipeptide composition of the protein sequence
        for amino acid type s and t (where s and t = 1,2,..,20), AA(s,t) is the number
        of dipeptides represented by amino acid type s and t and N is the total number
        of dipeptides.

        Returns
        -------
        dipeptide_comp_df : pd.DataFrame
            dataframe of dipeptide composition for all protein sequences. DataFrame will
            be of the shape N x 400, where N is the number of protein sequences and
            400 is the number of features calculated from the descriptor (20^2 for
            the 20 canonical amino acids).
        """
        print('\nGetting Dipeptide Composition Descriptors...')
        print('############################################\n')

        #if attribute already calculated & not empty then return it
        if not self.dipeptide_composition.empty:
            return self.dipeptide_composition

        dipeptide_comp = []

        #get feature names for DPComp descriptor
        dipeptides = list((AAComposition.CalculateDipeptideComposition(self.protein_seqs[0])).keys())

        #iterate through all sequences, calculate descriptor values and append to dipeptide_comp list
        for seq in tqdm(self.protein_seqs,unit=" sequences",position=0,desc="Dipeptide Composition", file=sys.stdout):
            AADipeptide = AAComposition.CalculateDipeptideComposition(seq)
            dipeptide_comp.append(list(AADipeptide.values()))

        #convert calculated DPComp values into dataframe
        dipeptide_comp_df = pd.DataFrame(data=dipeptide_comp, columns=dipeptides)

        self.dipeptide_composition = dipeptide_comp_df  #set descriptor attribute

        return dipeptide_comp_df

    def get_tripeptide_composition(self):
        """
        Calculate Tripeptide Composition (TPComp) of protein sequences.
        Tripeptide composition is the fraction of each tripeptide type within a
        protein sequence. With tripeptides being of length 3 and there being 20
        canonical amino acids this creates 20^3 different combinations, thus a
        8000-Dimensional vector will be produced such that:

        TPComp(s,t,u) = AA(s,t,u) / N -1

        where TPComp(s,t,u) is the tripeptide composition of the protein sequence
        for amino acid type s, t and u (where s, t and u = 1,2,..,20), AA(s,t,u) is
        the number of tripeptides represented by amino acid type s and t, and N is
        the total number of tripeptides.

        Returns
        -------
        tripeptide_df : pd.DataFrame
            dataframe of tripeptide composition for all protein sequences. DataFrame will
            be of the shape N x 8000, where N is the number of protein sequences and
            8000 is the number of features calculated from the descriptor (20^3 for
            the 20 canonical amino acids).
        """
        print('\nGetting Tripeptide Composition Descriptors...')
        print('#############################################\n')

        #if attribute already calculated & not empty then return it
        if not self.tripeptide_composition.empty:
            return self.tripeptide_composition

        tripeptide_comp = []

        #get feature names for TPComp descriptor
        tripeptides = list((AAComposition.GetSpectrumDict(self.protein_seqs[0])).keys())

        #iterate through all sequences, calculate descriptor values and append to tripeptide_comp list
        for seq in tqdm(self.protein_seqs,unit=" sequences",position=0,desc="Tripeptide Comp", file=sys.stdout):
            AATripeptide=AAComposition.GetSpectrumDict(seq)
            tripeptide_comp.append(list(AATripeptide.values()))

        #convert calculated TPComp values into dataframe
        tripeptide_comp_df = pd.DataFrame(data=tripeptide_comp, columns=tripeptides)

        self.tripeptide_composition = tripeptide_comp_df    #set descriptor attribute

        return tripeptide_comp_df

    def get_norm_moreaubroto_autocorrelation(
        self,aa_encoding=Autocorrelation._AAProperty,
        aa_name=Autocorrelation._AAPropertyName
        ):
        """
        Calculate Normalized MoreauBrotoAuto Autocorrelation (NMBAuto) descriptor.
        Autocorrelation descriptors are a class of topological descriptors,
        also known as molecular connectivity indices, that describe the level of
        correlation between two objects (protein or peptide sequences) in terms of
        their specific structural or physicochemical property [4], which are
        defined based on the distribution of amino acid properties along the sequence.
        8 amino acid properties are used for deriving the descriptors. The derivations
        and detailed explanations of this type of descriptor is outlind in [5].
        Each autocorrelation will generate 240 features, using the default 8 properties:

        AccNo. CIDH920105 - Normalized Average Hydrophobicity Scales
        AccNo. BHAR880101 - Average Flexibility Indices
        AccNo. CHAM820101 - Polarizability Parameter
        AccNo. CHAM820102 - Free Energy of Solution in Water, kcal/mole
        AccNo. CHOC760101 - Residue Accessible Surface Area in Tripeptide
        AccNo. BIGC670101 - Residue Volume
        AccNo. CHAM810101 - Steric Parameter
        AccNo. DAYM780201 - Relative Mutability

        The NMBAuto descriptor is a type of Autocorrelation descriptor that uses
        the property values as the basis for measurement [5].

        Parameters
        ----------
        aa_encoding : list of dicts (default = Autocorrelation._AAProperty)
            list of dictionaries of amino acid values for particular properties,
            in the form "A": 0.2, "B":0.4 etc. By default 8 properties are used,
            _AAProperty = (_Hydrophobicity,_AvFlexibility,_Polarizability,
            _FreeEnergy,_ResidueASA,_ResidueVol,_Steric,_Mutability).
        aa_name : list of str (default = Autocorrelation._AAPropertyName)
            list of strings containing the name of each property. Used for the
            column/feature names in the output dataframe. Default is just a list
            of string representations of each of the mentioned 8 properties.

        Returns
        -------
        norm_moreaubroto_autocorr_df : pd.DataFrame
            dataframe of NMBAuto values for all protein sequences. DataFrame will
            be of the shape N x 240, where N is the number of protein sequences and
            240 is the number of features calculated from the descriptor (30 features
            per property - using 8 properties.)
        """
        print('\nGetting Normalized Moreaubroto Autocorrelation Descriptors...')
        print('#############################################################\n')

        #if attribute already calculated & not empty then return it
        if not self.normalized_moreaubroto_autocorrelation.empty:
            return self.normalized_moreaubroto_autocorrelation

        #initialise lists used to store the autocorrelation values and keys
        norm_moreaubroto_autocorr = []
        norm_moreaubroto_autocorr_keys = []
        norm_moreaubroto_autocorr_values = []

        #iterate through protein sequences and calculate NMBAuto values, append to list
        for seq in tqdm(self.protein_seqs,unit=" sequences",position=0,desc="NMBAuto", file=sys.stdout):
            norm_moreau_broto = Autocorrelation.CalculateNormalizedMoreauBrotoAuto(seq, aa_encoding,aa_name)
            norm_moreaubroto_autocorr.append(norm_moreau_broto)

        #get keys/column names generated from the descriptor calculation
        for name in range(0,len(aa_name)):
          norm_moreaubroto_autocorr_keys.append(list(norm_moreaubroto_autocorr[0][aa_name[name]].keys()))

        #flatten column names/keys into 1 1-D list
        norm_moreaubroto_autocorr_keys = flatten(norm_moreaubroto_autocorr_keys)

        '''
        Reformatting the output of the descriptors from the PyBioMed package.
        The PyBioMed descriptor function returns a dict of dicts of the resultant
        values. For uniformity with the rest of the descriptor calculations,
        this output is transformed such that it can be output into a dataframe.
        '''
        for x in range(0,len(list(norm_moreaubroto_autocorr))):
            for y in range(0,len(list(norm_moreaubroto_autocorr[x].values()))):
                norm_moreaubroto_autocorr_values.append(
                    list((list((list(norm_moreaubroto_autocorr)[x]).values()))[y].values())
                )

        #flatten list of lists of the autocorrelation values
        norm_moreaubroto_autocorr_values = flatten(norm_moreaubroto_autocorr_values)

        #reshape into N x M array where N is the number of protein sequences and M
        #   is the number of AA properties used in the correlation
        norm_moreaubroto_autocorr_values_ = np.reshape(norm_moreaubroto_autocorr_values,
            (len(self.protein_seqs),len(norm_moreaubroto_autocorr_keys)))

        #convert NMBAuto array values into a dataframe
        norm_moreaubroto_autocorr_df = pd.DataFrame(data=norm_moreaubroto_autocorr_values_,
            columns=norm_moreaubroto_autocorr_keys)

        self.normalized_moreaubroto_autocorrelation = norm_moreaubroto_autocorr_df  #set descriptor attribute

        return norm_moreaubroto_autocorr_df

    def get_moran_autocorrelation(
        self,aa_encoding=Autocorrelation._AAProperty,
        aa_name=Autocorrelation._AAPropertyName
        ):
        """
        **refer to NMBAuto docstring for autocorrelation description.
        Moran autocorrelation (MAuto) utilizes property deviations from the
        average values [5].

        Parameters
        ----------
        aa_encoding : list of dicts (default = Autocorrelation._AAProperty)
            **refer to NMBAuto doctring.
        aa_name : list of str (default = Autocorrelation._AAPropertyName)
            **refer to NMBAuto doctring.

        Returns
        -------
        moran_autocorr_df : pd.DataFrame
            dataframe of MAuto values for all protein sequences. DataFrame will
            be of the shape N x 240, where N is the number of protein sequences and
            240 is the number of features calculated from the descriptor (30 features
            per property - using 8 properties.)
        """
        print('\nGetting Moran Autocorrelation Descriptors...')
        print('############################################\n')

        #if attribute already calculated & not empty then return it
        if not self.moran_autocorrelation.empty:
            return self.moran_autocorrelation

        #initialise lists used to store the MAuto values and keys
        moran_autocorr = []
        moran_autocorr_keys = []
        moran_autocorr_values = []

        #iterate through protein sequences and calculate MAuto values, append to list
        for seq in tqdm(self.protein_seqs,unit=" sequences",position=0,desc="MAuto", file=sys.stdout):
            moranautocorr = Autocorrelation.CalculateMoranAuto(seq, aa_encoding,aa_name)
            moran_autocorr.append(moranautocorr)

        #get keys/column names generated from the descriptor calculation
        for name in range(0,len(aa_name)):
          moran_autocorr_keys.append(list(moran_autocorr[0][aa_name[name]].keys()))

        #flatten column names/keys into 1 1-D list
        moran_autocorr_keys = flatten(moran_autocorr_keys)

        '''
        Reformatting the output of the descriptors from the PyBioMed package.
        The PyBioMed descriptor function returns a dict of dicts of the resultant
        values. For uniformity with the rest of the descriptor calculations,
        this output is transformed such that it can be output into a dataframe.
        '''
        for x in range(0,len(list(moran_autocorr))):
            for y in range(0,len(list(moran_autocorr[x].values()))):
                moran_autocorr_values.append(
                    list((list((list(moran_autocorr)[x]).values()))[y].values())
                )

        #flatten list of lists of the autocorrelation values
        moran_autocorr_values = flatten(moran_autocorr_values)

        #reshape into N x M array where N is the number of protein sequences and M
        #   is the number of AA properties used in the correlation
        moran_autocorr_values_ = np.reshape(moran_autocorr_values,
            (len(self.protein_seqs),len(moran_autocorr_keys)))

        #convert MAuto array values into a dataframe
        moran_autocorr_df = pd.DataFrame(data=moran_autocorr_values_, columns=moran_autocorr_keys)

        self.moran_autocorrelation = moran_autocorr_df  #set descriptor attribute

        return moran_autocorr_df

    def get_geary_autocorrelation(
        self,aa_encoding=Autocorrelation._AAProperty,
        aa_name=Autocorrelation._AAPropertyName
        ):
        """
        *refer to NMBAuto docstring for autocorrelation description.
        Geary Autocorrelation (GAuto) utilizes the square-difference of property
        values instead of vector-products (of property values or deviations)[5].

        Parameters
        ----------
        aa_encoding : list of dicts (default = Autocorrelation._AAProperty)
            **refer to NMBAuto doctring.
        aa_name : list of str (default = Autocorrelation._AAPropertyName)
            **refer to NMBAuto doctring.

        Returns
        -------
        geary_autocorr_df : pd.DataFrame
            dataframe of GAuto values for all protein sequences. DataFrame will
            be of the shape N x 240, where N is the number of protein sequences and
            240 is the number of features calculated from the descriptor (30 features
            per property - using 8 properties.)
        """
        print('\nGetting Geary Autocorrelation Descriptors...')
        print('############################################\n')

        #if attribute already calculated & not empty then return it
        if not self.geary_autocorrelation.empty:
            return self.geary_autocorrelation

        #initialise lists used to store the autocorrelation values and keys
        geary_autocorr = []
        geary_autocorr_keys = []
        geary_autocorr_values = []

        #iterate through protein sequences and calculate GAuto values, append to list
        for seq in tqdm(self.protein_seqs,unit=" sequences",position=0,desc="GAuto", file=sys.stdout):
            gearyautocorr = Autocorrelation.CalculateGearyAuto(seq, aa_encoding,aa_name)
            geary_autocorr.append(gearyautocorr)

        #get keys/column names generated from the descriptor calculation
        for name in range(0,len(aa_name)):
          geary_autocorr_keys.append(list(geary_autocorr[0][aa_name[name]].keys()))

        #flatten column names/keys into 1 1-D list
        geary_autocorr_keys = flatten(geary_autocorr_keys)

        '''
        Reformatting the output of the descriptors from the PyBioMed package.
        The PyBioMed descriptor function returns a dict of dicts of the resultant
        values. For uniformity with the rest of the descriptor calculations,
        this output is transformed such that it can be output into a dataframe.
        '''
        for x in range(0,len(list(geary_autocorr))):
            for y in range(0,len(list(geary_autocorr[x].values()))):
                geary_autocorr_values.append(
                    list((list((list(geary_autocorr)[x]).values()))[y].values())
                )

        #flatten list of lists of the autocorrelation values
        geary_autocorr_values = flatten(geary_autocorr_values)

        #reshape into N x M array where N is the number of protein sequences and M
        #   is the number of AA properties used in the correlation
        geary_autocorr_values_ = np.reshape(geary_autocorr_values, (len(self.protein_seqs),len(geary_autocorr_keys)))

        #convert GAuto array values into a dataframe
        geary_autocorr_df = pd.DataFrame(data=geary_autocorr_values_, columns=geary_autocorr_keys)

        self.geary_autocorrelation = geary_autocorr_df  #set descriptor attribute

        return geary_autocorr_df

    def get_ctd(self):
        """
        Calculate Composition, transition and distribution (CTD) features of protein sequences.
        Composition is the number of amino acids of a particular property (e.g., hydrophobicity)
        divided by the total number of amino acids in a protein sequence. Transition
        characterizes the percent frequency with which amino acids of a particular
        property is followed by amino acids of a different property. Distribution
        measures the chain length within which the first, 25%, 50%, 75%, and 100% of
        the amino acids of a particular property are located, respectively [6].
        CTD functionality in the PyBioMed package uses the properties:
        Polarizability, Solvent Accessibility, Secondary Structure, Charge,
        Polarity, Normalized VDWV, Hydrophobicity. The output will be of shape
        N x 147 where N is the number of protein sequences. 21/147 will be
        composition, 21/147 will be transition and the remaining 105 are distribution.

        Returns
        -------
        ctd_df : pd.DataFrame
            dataframe of CTD descriptor values for all protein sequences. DataFrame will
            be of the shape N x 147, where N is the number of protein sequences and
            147 is the number of features calculated from the descriptors.
        """
        print('\nGetting CTD Descriptors...')
        print('##########################\n')

        #if attribute already calculated & not empty then return it
        if not self.ctd.empty:
            return self.ctd

        ctd = []

        #get feature names for CTD descriptor
        keys = list((CTD.CalculateCTD(self.protein_seqs[0])).keys())

        #iterate through sequences, calculating the CTD values using PyBioMed Package
        for seq in tqdm(self.protein_seqs,unit=" sequences",position=0,desc="CTD", file=sys.stdout):
            ctd_ = CTD.CalculateCTD(seq)
            ctd.append(list(ctd_.values()))

        #convert calculated ctd values into dataframe
        ctd_df = pd.DataFrame(data=ctd, columns=keys)

        self.ctd = ctd_df                       #set descriptor attributes
        self.composition = ctd_df.iloc[:,:21]
        self.transition = ctd_df.iloc[:, 21: 42]
        self.distribution = ctd_df.iloc[:, 42: ]

        return ctd_df

    def get_conjoint_triad(self):
        """
        Calculate Conjoint Triad features (CTriad) for the protein sequences. CTF
        mainly considers neighbor relationships in protein sequences by encoding
        each protein sequence using the triad (continuous three amino acids)
        frequency distribution extracted from a 7-letter reduced alphabet [7]. This
        descriptor calculates 343 different features (7x7x7), with the output
        being of shape N x 343 where N is the number of sequences.

        Returns
        -------
        ct_df : pd.DataFrame
            dataframe of CTriad descriptor values for all protein sequences. DataFrame
            will be of the shape N x 343, where N is the number of protein sequences and
            343 is the number of features calculated from the descriptor.
        """
        print('\nGetting Conjoint Triad Descriptors...')
        print('#####################################\n')

        #if attribute already calculated & not empty then return it
        if not self.conjoint_triad.empty:
            return self.conjoint_triad

        ct = []

        #iterate through sequences, calculating the CTriad values using PyBioMed Package
        for seq in tqdm(self.protein_seqs,unit=" sequences",position=0,desc="Conjoint Triad", file=sys.stdout):
            con_Tri = {}
            protein_num = ConjointTriad._Str2Num(seq)
            for i in range(7):    #manually calculating CTriad features due to error in PyProtein
                for j in range(7):
                    for k in range(7):
                        temp = str(i) + str(j) + str(k)
                        con_Tri[temp] = protein_num.count(temp)
            ct.append(list(con_Tri.values()))

        #get feature/column names for CTriad descriptor
        keys = list(con_Tri.keys())

        #convert calculated CTriad values into dataframe
        ct_df = pd.DataFrame(data=ct, columns=keys)

        self.conjoint_triad = ct_df             #set descriptor attribute

        return ct_df

    def get_seq_order_coupling_number(self, maxlag=30):
        """
        Calculate Sequence Order Coupling Number (SOCNum) features for the protein sequences.
        Sequence Order Coupling Number computes the dissimilarity between amino acid
        pairs. The distance between amino acid pairs is determined by d which varies
        between 1 to nlag. For each d, it computes the sum of the dissimilarities
        of all amino pairs [8]. The number of output features depends on the value
        of maxlag, by default this value is 30 so 60 features are output.

        Parameters
        ----------
        maxlag : int (default = 30)
            maxlag is the maximum lag and the length of the protein should be larger
            than maxlag. Default set to 30.

        Returns
        -------
        seq_order_df : pd.DataFrame
            dataframe of SOCNum descriptor values for all protein sequences. DataFrame
            will be of the shape N x 60, where N is the number of protein sequences and
            60 is the number of features calculated from the descriptor.
        """
        print('\nGetting Sequence Order Coupling Descriptors...')
        print('##############################################\n')

        #if attribute already calculated & not empty then return it
        if not self.seq_order_coupling_number.empty:
            return self.seq_order_coupling_number

        seq_order = []

        #iterate through sequences, calculating the SOCNum values using PyBioMed Package
        for seq in tqdm(self.protein_seqs,unit=" sequences",position=0,desc="Sequence Order Coupling Number", file=sys.stdout):
            seq_order_num = QuasiSequenceOrder.GetSequenceOrderCouplingNumberTotal(seq,maxlag=maxlag)
            seq_order.append(seq_order_num)

        #convert calculated SOCNum values into dataframe
        seq_order_df = pd.DataFrame(data=seq_order)

        self.seq_order_coupling_number = seq_order_df   #set descriptor attribute

        return seq_order_df

    def get_quasi_seq_order(self, maxlag=30, weight=0.1):
        """
        Calculate Quasi Sequence Order features for the protein sequences.
        The quasi-sequence-order descriptors were proposed by K.C. Chou, et.al. [8]
        They are derived from the distance matrix between the 20 amino acids. The
        Scheider-Wrede physicochemical distance matrix and Grantham chemical
        distance matrix used by Grantham et. al. are both used in the calculation
        of the descriptor. 100 values are calculated per sequence, thus generating
        an output of N x 100 where N is the numnber of sequences.

        Parameters
        ----------
        maxlag : int (default = 30)
            **refer to seq-order-coupling-number docstring.
        weight : float (default = 0.1)
            weight factor, can select any value within the region from 0.05 to 0.7.

        Returns
        -------
        quasi_seq_order_df : pd.DataFrame
            dataframe of quasi-sequence-order descriptor values for the
            protein sequences, with output shape N x 100 where N is the number
            of sequences and 100 the number of calculated features.
        """
        print('\nGetting Quasi Sequence Order Descriptors...')
        print('###########################################\n')

        #if attribute already calculated & not empty then return it
        if not self.quasi_seq_order.empty:
            return self.quasi_seq_order

        #if maxlag greater than length of protein, set to default value of 30
        if (maxlag >= len(self.protein_seqs[0])):
            maxlag=30

        quasi_seq_order = []

        #get feature names for QuasiSeqOrder descriptor
        keys = list((QuasiSequenceOrder.GetQuasiSequenceOrder(self.protein_seqs[0])).keys())

        #iterate through sequences, calculating the descriptor values using PyBioMed Package
        for seq in tqdm(self.protein_seqs,unit=" sequences",position=0,desc="Quasi Sequence Order",file=sys.stdout):
            quasi_seq = QuasiSequenceOrder.GetQuasiSequenceOrder(seq,maxlag=maxlag, weight=weight)
            quasi_seq_order.append(list(quasi_seq.values()))

        #convert calculated quasi_seq_order values into dataframe
        quasi_seq_order_df = pd.DataFrame(data=quasi_seq_order, columns=keys)

        self.quasi_seq_order = quasi_seq_order_df    #set descriptor attribute

        return quasi_seq_order_df

    def get_pseudo_aa_composition(
            self, lambda_=30, weight=0.05, AAP=[PseudoAAC._Hydrophobicity,
            PseudoAAC._hydrophilicity, PseudoAAC._residuemass]
            ):
        """
        Calculate Pseudo Amino Acid Composition features for the protein sequences.
        Similar to the quasi-sequence order descriptor, the pseudo amino acid descriptor is
        made up of a 50-dimensional vector in which the first 20 components reflect the
        effect of the amino acid composition and the remaining 30 components reflect
        the effect of sequence order [9]. 50 values will be calculated per sequence,
        thus generating an output of N x 50 where N is the number of sequences.

        Parameters
        ----------
        lambda_ : int (default = 30)
            Lambda factor reflects the rank of correlation and is a non-negative integer.
            The value should not be larger than the length of input protein
            sequences. When lambda=0, the output is the 20-D amino acid composition.
        weight : float (default = 0.05)
            weight factor is designed for the users to put weight on the additional
            pseudo AAC components with respect to the conventional AA components.
            The user can select any value within the region from 0.05 to 0.7.
        AAP : list (default = AAP=[PseudoAAC._Hydrophobicity,
                PseudoAAC._hydrophilicity, PseudoAAC._residuemass])
            list contraining various physiochemical properties of the sequences.
            Properites will be a list of dicts, in the form:
            'Property_Name': Value (float). By default, 3 properties are used:
            hydrophobicity, hydrophillicity and residue mass of the amino acids.
            Refer to PyProtein module in PyBioMed package for more info:
            https://github.com/gadsbyfly/PyBioMed/tree/master/PyBioMed/PyProtein

        Returns
        -------
        pseudo_aa_composition_df : pd.DataFrame
            dataframe of pseudo amino acid composition descriptor values
            for the protein sequences of output shape N x 50.
        """
        print('\nGetting Pseudo Amino Acid Composition Descriptors...')
        print('####################################################\n')

        #if attribute already calculated & not empty then return it
        if not self.pseudo_aa_composition.empty:
            return self.pseudo_aa_composition

        #if lambda not a non-negative int then set to default value of 30
        if not (lambda_>=0):
            lambda_ = 30

        #if weight not b/w 0.05 and 0.7 then set to default value of 0.05
        if (weight<0.05 or weight>0.7):
            weight = 0.05

        pseudo_AA_Comp = []

        #get feature names for pseudoAAComp descriptor
        keys = (list(PseudoAAC.GetPseudoAAC(self.protein_seqs[0], AAP=AAP).keys()))

        #iterate through sequences, calculating the descriptor values using PyBioMed Package
        for seq in tqdm(self.protein_seqs,unit=" sequences",position=0,desc="Pseudo Amino Acid Composition",file=sys.stdout):
            pseudo_AA=PseudoAAC.GetPseudoAAC(seq, lamda=lambda_, weight=weight, AAP=AAP)
            pseudo_AA_Comp.append(list(pseudo_AA.values()))

        #convert calculated PseudoAAComp values into dataframe
        pseudo_aa_composition_df = pd.DataFrame(data=pseudo_AA_Comp, columns=keys)

        self.pseudo_aa_composition = pseudo_aa_composition_df      #set descriptor attribute

        return pseudo_aa_composition_df

    def get_amp_pseudo_aa_composition(self, lambda_=30, weight=0.5):
        """
        Calculate Amphiphilic (Type II) Pseudo amino acid composition features
        for the protein sequences. Amphiphilic pseudo amino acid composition has
        the same form as the amino acid compositon, but contains much more
        information that is related to the sequence order of a protein and the
        distribution of the hydrophobic and hydrophilic amino acids along its chain [10].
        80 descriptor values will be generated, producing an output of the shape
        N x 80 where N is the number of sequences and 80 is the number of features
        calculated from descriptor.

        Parameters
        ----------
        lambda_ : int (default = 30)
            **refer to Pseudo Amino Acid Composition docstring
        weight : float (default = 0.5)
            **refer to Pseudo Amino Acid Composition docstring

        Returns
        -------
        amp_pseudo_aa_composition_df : pd.DataFrame
            dataframe of Amphiphilic pseudo amino acid composition descriptor
            values for the protein sequences of output shape N x 80.
        """
        print('\nGetting Amphiphilic Amino Acid Composition Descriptors...')
        print('#########################################################\n')

        #if attribute already calculated & not empty then return it
        if not self.amp_pseudo_aa_composition.empty:
            return self.amp_pseudo_aa_composition

        #if lambda not a non-negative int then set to default value of 30
        if not (lambda_>=0):
            lambda_ = 30

        #if weight not b/w 0.05 and 0.7 then set to default value of 0.05
        if (weight<0.05 or weight>0.7):
            weight = 0.05

        amp_pseudo_AAComp = []

        #get column names/keys for descriptor
        keys = list((PseudoAAC.GetAPseudoAAC(self.protein_seqs[0])).keys())

        #iterate through sequences, calculating the descriptor values using PyBioMed Package
        for seq in tqdm(self.protein_seqs,unit=" sequences",position=0,desc="Ampiphillic Amino Acid Composition",file=sys.stdout):
            amp_pseudo = PseudoAAC.GetAPseudoAAC(seq, lamda=lambda_, weight=weight)
            amp_pseudo_AAComp.append(list(amp_pseudo.values()))

        #convert calculated Ampiphillic pseudo_AAComp values into dataframe
        amp_pseudo_aa_composition_df = pd.DataFrame(data=amp_pseudo_AAComp, columns=keys)

        self.amp_pseudo_aa_composition = amp_pseudo_aa_composition_df  #set descriptor attribute

        return amp_pseudo_aa_composition_df

    def get_descriptor_encoding(self,descriptor):
        """
        Get the protein descriptor values of a specified input descriptor. If the
        sought descriptor has already been calculated then its attribute is returned,
        else the descriptor is calculated using its get_descriptor function.

        Parameters
        ----------
        descriptor : str
            name of descriptor to return. Method can accept the approximate name
            of the descriptor, e.g. 'aa_comp'/'aa_compo' etc will return the
            'aa_composition' descriptor. This functionality is realised using
            the difflib library.

        Returns
        -------
        desc_encoding : pd.DataFrame/None
            dataframe of descriptor values, calculated from the descriptor
            instances' protein sequences. None returned if no matching descriptor found.
        """
        #validate input descriptor is a valid available descriptor, get its closest match
        desc_matches = get_close_matches(descriptor,self.valid_descriptors(),cutoff=0.4)
        if desc_matches!=[]:
            desc = desc_matches[0]  #set desc to closest descriptor match found
        else:
            raise ValueError('Could not find a match for the input descriptor {} \
                in available valid models: {}.'.format(descriptor, self.valid_descriptors()))

        #if sought descriptor attribute dataframe is empty, call the descriptor's
        #   get_descriptor() function, set desc_encoding to descriptor attribute
        if desc == 'aa_composition':
            if (getattr(self, desc).empty):
                self.get_aa_composition()
            desc_encoding = self.aa_composition
        elif desc == 'dipeptide_composition':
            if (getattr(self, desc).empty):
                self.get_dipeptide_composition()
            desc_encoding = self.dipeptide_composition
        elif desc == 'tripeptide_composition':
            if (getattr(self, desc).empty):
                self.get_tripeptide_composition()
            desc_encoding = self.tripeptide_composition
        elif desc == 'normalized_moreaubroto_autocorrelation':
            if (getattr(self, desc).empty):
              self.get_norm_moreaubroto_autocorrelation()
            desc_encoding = self.normalized_moreaubroto_autocorrelation
        elif desc == 'moran_autocorrelation':
            if (getattr(self, desc).empty):
              self.get_moran_autocorrelation()
            desc_encoding = self.moran_autocorrelation
        elif desc == 'geary_autocorrelation':
            if (getattr(self, desc).empty):
              self.get_geary_autocorrelation()
            desc_encoding = self.geary_autocorrelation
        elif desc == 'ctd':
            if (getattr(self, desc).empty):
              self.get_ctd()
            desc_encoding = self.ctd
        elif desc == 'composition':
            if (getattr(self, desc).empty):
              self.get_ctd()
            desc_encoding = self.composition
        elif desc == 'transition':
            if (getattr(self, desc).empty):
              self.get_ctd()
            desc_encoding = self.transition
        elif desc == 'distribution':
            if (getattr(self, desc).empty):
              self.get_ctd()
            desc_encoding = self.distribution
        elif desc == 'conjoint_triad':
            if (getattr(self, desc).empty):
              self.get_conjoint_triad()
            desc_encoding = self.conjoint_triad
        elif desc == 'seq_order_coupling_number':
            if (getattr(self, desc).empty):
              self.get_seq_order_coupling_number()
            desc_encoding = self.seq_order_coupling_number
        elif desc == 'quasi_seq_order':
            if (getattr(self, desc).empty):
              self.get_quasi_seq_order()
            desc_encoding = self.quasi_seq_order
        elif desc == 'pseudo_aa_composition':
            if (getattr(self, desc).empty):
              self.get_pseudo_aa_composition()
            desc_encoding = self.pseudo_aa_composition
        elif desc == 'amp_pseudo_aa_composition':
            if (getattr(self, desc).empty):
              self.get_amp_pseudo_aa_composition()
            desc_encoding = self.amp_pseudo_aa_composition
        else:
          desc_encoding = None           #no matching descriptor

        return desc_encoding

    def all_descriptors_list(self, desc_combo=1):
       """
       Get list of all available descriptor attributes. Using the desc_combo
       input parameter you can get the list of all descriptors, all combinations
       of 2 descriptors or all combinations of 3 descriptors. Default of 1 will
       mean a list of all available descriptor attributes will be returned.

       Parameters
       ----------
       desc_combo : int (default = 1)
            combination of descriptors to return. A value of 2 or 3 will return
            all combinations of 2 or 3 descriptor attributes etc.

       Returns
       -------
       all_descriptors : list
            list of available descriptor attributes.
       """
       #filter out class attributes that are not any of the desired descriptors
       all_descriptors = list(filter(lambda x: x.startswith('_'), list(self.__dict__.keys())))
       all_descriptors = list(filter(lambda x: not x.startswith('_all_desc'), all_descriptors))
       all_descriptors = [de[1:] for de in all_descriptors]

       #get all combinations of 2 or 3 descriptors.
       if desc_combo == 2:
           all_descriptors = list(itertools.combinations(all_descriptors, 2))
       elif desc_combo == 3:
           all_descriptors = list(itertools.combinations(all_descriptors, 3))
       else:
           pass     #if desc_combo not equal to 2 or 3 then use default all_descriptors

       return all_descriptors

    def get_all_descriptors(self):
        """
        Calculate all individual descriptor values, concatenating each descriptor
        dataframe into one storing all descriptors.

        Returns
        -------
        all_desc_df : pd.DataFrame
            concatenated dataframe of all individual descriptors. Output will be
            of the shape N x 9920, where N is the number of sequences and
            9920 is the total number of descriptor features calculated.
        """
        print('Calculating all Descriptor values....\n')
        print('#####################################\n')

        #if descriptor attribute DF is empty then call get_descriptor function
        if (getattr(self, "aa_composition").empty):
                self.aa_composition = self.get_aa_composition()

        if (getattr(self, "dipeptide_composition").empty):
                self.dipeptide_composition = self.get_dipeptide_composition()

        if (getattr(self, "tripeptide_composition").empty):
                self.tripeptide_composition = self.get_tripeptide_composition()

        if (getattr(self, "normalized_moreaubroto_autocorrelation").empty):
            self.normalized_moreaubroto_autocorrelation = self.get_norm_moreaubroto_autocorrelation()

        if (getattr(self, "moran_autocorrelation").empty):
            self.moran_autocorrelation = self.get_moran_autocorrelation()

        if (getattr(self, "geary_autocorrelation").empty):
            self.geary_autocorrelation = self.get_geary_autocorrelation()

        if (getattr(self, "ctd").empty):
                self.ctd = self.get_ctd()

        if (getattr(self, "conjoint_triad").empty):
                self.conjoint_triad = self.get_conjoint_triad()

        if (getattr(self, "seq_order_coupling_number").empty):
                self.seq_order_coupling_number = self.get_seq_order_coupling_number()

        if (getattr(self, "quasi_seq_order").empty):
                self.quasi_seq_order = self.get_quasi_seq_order()

        if (getattr(self, "pseudo_aa_composition").empty):
                self.pseudo_aa_composition = self.get_pseudo_aa_composition()

        if (getattr(self, "amp_pseudo_aa_composition").empty):
                self.amp_pseudo_aa_composition = self.get_amp_pseudo_aa_composition()

        #append all calculated descriptors to list
        all_desc = [
            self.aa_composition, self.dipeptide_composition, self.tripeptide_composition,
            self.normalized_moreaubroto_autocorrelation, self.moran_autocorrelation,
            self.geary_autocorrelation, self.composition, self.transition,
            self.distribution, self.ctd, self.conjoint_triad, self.seq_order_coupling_number,
            self.quasi_seq_order, self.pseudo_aa_composition, self.amp_pseudo_aa_composition
            ]

        #concatenate individual descriptor dataframe attributes
        all_desc_df = pd.concat(all_desc, axis = 1)

        self.all_descriptors = all_desc_df      #set all_descriptors attribute

        return all_desc_df

    def valid_descriptors(self):
        """
        Get a list of all valid descriptors available in the module.

        Returns
        -------
        valid_desc : list
            list of all valid descriptors that the module supports.
        """
        valid_desc = [
        'aa_composition', 'dipeptide_composition', 'tripeptide_composition',
        'normalized_moreaubroto_autocorrelation','moran_autocorrelation','geary_autocorrelation',
        'ctd', 'composition', 'transition', 'distribution', 'conjoint_triad',
        'seq_order_coupling_number','quasi_seq_order',
        'pseudo_aa_composition', 'amp_pseudo_aa_composition'
        ]
        return valid_desc

######################          Getters & Setters          ######################

    @property
    def all_desc(self):
        return self._all_desc

    @all_desc.setter
    def all_desc(self, val):
        self._all_desc = val

    @property
    def aa_composition(self):
        return self._aa_composition

    @aa_composition.setter
    def aa_composition(self, val):
        self._aa_composition = val

    @property
    def dipeptide_composition(self):
        return self._dipeptide_composition

    @dipeptide_composition.setter
    def dipeptide_composition(self, val):
        self._dipeptide_composition = val

    @property
    def tripeptide_composition(self):
        return self._tripeptide_composition

    @tripeptide_composition.setter
    def tripeptide_composition(self, val):
        self._tripeptide_composition = val

    @property
    def normalized_moreaubroto_autocorrelation(self):
        return self._normalized_moreaubroto_autocorrelation

    @normalized_moreaubroto_autocorrelation.setter
    def normalized_moreaubroto_autocorrelation(self, val):
        self._normalized_moreaubroto_autocorrelation = val

    @property
    def moran_autocorrelation(self):
        return self._moran_autocorrelation

    @moran_autocorrelation.setter
    def moran_autocorrelation(self, val):
        self._moran_autocorrelation = val

    @property
    def geary_autocorrelation(self):
        return self._geary_autocorrelation

    @geary_autocorrelation.setter
    def geary_autocorrelation(self, val):
        self._geary_autocorrelation = val

    @property
    def ctd(self):
        return self._ctd

    @ctd.setter
    def ctd(self, val):
        self._ctd = val

    @property
    def composition(self):
        return self._composition

    @composition.setter
    def composition(self, val):
        self._composition = val

    @property
    def transition(self):
        return self._transition

    @transition.setter
    def transition(self, val):
        self._transition = val

    @property
    def distribution(self):
        return self._distribution

    @distribution.setter
    def distribution(self, val):
        self._distribution = val

    @property
    def conjoint_triad(self):
        return self._conjoint_triad

    @conjoint_triad.setter
    def conjoint_triad(self, val):
        self._conjoint_triad = val

    @property
    def seq_order_coupling_number(self):
        return self._seq_order_coupling_number

    @seq_order_coupling_number.setter
    def seq_order_coupling_number(self, val):
        self._seq_order_coupling_number = val

    @property
    def quasi_seq_order(self):
        return self._quasi_seq_order

    @quasi_seq_order.setter
    def quasi_seq_order(self, val):
        self._quasi_seq_order = val

    @property
    def pseudo_aa_composition(self):
        return self._pseudo_aa_composition

    @pseudo_aa_composition.setter
    def pseudo_aa_composition(self, val):
        self._pseudo_aa_composition = val

    @property
    def amp_pseudo_aa_composition(self):
        return self._amp_pseudo_aa_composition

    @amp_pseudo_aa_composition.setter
    def amp_pseudo_aa_composition(self, val):
        self._amp_pseudo_aa_composition = val

    @property
    def all_descriptors(self):
        return self._all_descriptors

    @all_descriptors.setter
    def all_descriptors(self, val):
        self._all_descriptors = val

    @all_descriptors.deleter
    def all_descriptors(self):
        """ Delete all descriptor attribute dataframes """
        del self._all_descriptors
        del self._aa_composition
        del self._dipeptide_composition
        del self._tripeptide_composition
        del self._normalized_moreaubroto_autocorrelation
        del self._moran_autocorrelation
        del self._geary_autocorrelation
        del self._ctd
        del self._transition
        del self._composition
        del self._distribution
        del self._conjoint_triad
        del self._seq_order_coupling_number
        del self._quasi_seq_order
        del self._pseudo_aa_composition
        del self._amp_pseudo_aa_composition

################################################################################

    def __str__(self):
        return "Descriptor(Num Sequences: {}, Using All Descriptors: {})".format(
            self.num_seqs, self.all_desc)

    def __repr__(self):
        return ('<Descriptor: {}>'.format(self))

    def __len__(self):
        return len(self.all_descriptors)

    def __shape__(self):
        return self.all_descriptors.shape



###############################################################################
#####   Setup.py - installs all the required packages and dependancies    #####
###############################################################################

import pathlib
from setuptools import setup, find_packages
import sys

#ensure python version is greater than 3
if sys.version_info[0] < 3:
    sys.exit('Python 3 is the minimum version requirement')

HERE = pathlib.Path(__file__).parent

README = (HERE / 'README.md').read_text()

setup(name='pySAR',
# <<<<<<< HEAD
#       version='0.0.5',
# =======
#       version='0.0.5',
# >>>>>>> a0598ffeb0c057700cede56f2b509cc7110469cf
      version='0.0.9',
      description='A Python package used to analysis Protein Sequence Activity Relationships',
      long_description = README,
      long_description_content_type = "text/markdown",
      url = "https://github.com/amckenna41/pySAR",
      author='Adam McKenna',
      author_email='amckenna41@qub.ac.uk',
      license='MIT',
      classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
    ],
      install_requires=[
          'numpy>=1.16.6',
          'pandas',
          'scipy',
          'delayed',
          'scikit-learn==0.24.1',
          'requests',
          'matplotlib',
          'seaborn',
          'tqdm'
      ],
     # packages=find_packages(), #create Manifest file to ignore results folder in dist
     packages=find_packages(exclude=["Results"]),
     include_package_data=True,
     zip_safe=False)



# -*- coding: utf-8 -*-
"""
PyBioMed
========

	PyBioMed is a feature-rich packages used for the characterization of various complex biological molecules and interaction samples, such as chemicals, proteins, DNA, and their interactions. In order to be convenient to users, PyBioMed provides the module to get molecule structures, protein sequence and DNA sequence from Internet.

Using
-----

	Just write in Python

	>>> from PyBioMed.PyMolecule.estate import GetEstate
	>>> from rdkit import Chem
    >>> smi5 = 'COCCCC'
    >>> mol = Chem.MolFromSmiles(smi)
    >>> GetEstate(mol)

CBDD Group
==========

	CBDD Group（Computational Biology & Drug Design Group）was founded in 2012, it belongs to the Department of Medicinal Chemistry, School of Pharmaceutical Sciences of CSU. The center relies on a number of more powerful hardware platform (HP ProLiant DL980 G7 core computing cluster) and comprehensive software platform (MOE, IPA, Discovery Studio, Sybyl etc.). The center has carried out a series of studies in drug design, chemoinformatics and Artificial Intelligent Systems for Drug Discovery and Disease Diagnosis. Please visist the CBDD Group homepage http://home.scbdd.com/index.php?s=/Home/Index.html&t=english.

"""
